# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`~giza.inheritance` is a library for managing content reuse using an
inheritance-based model for structured content. In systems that use this code,
writers define content in a structured form, where content units can "inherit"
from other content units and optionally override some portions of the inherited
units.
"""

import copy
import collections
import logging
import os.path
import sys

import jinja2
import yaml

from giza.libgiza.config import RecursiveConfigurationBase, ConfigurationBase

logger = logging.getLogger('giza.libgiza.inheritance')

if sys.version_info >= (3, 0):
    basestring = str


class InheritableContentError(Exception):
    """
    Exception used by inheritance code to indicate a problem resolving
    inherited content.
    """

    pass


class InheritanceReference(RecursiveConfigurationBase):
    """
    Represents a single reference to another unit of content. The
    setter method for the :meth:`~giza.inheritance.InheritanceReference.file`
    attribute, returns an error if it cannot discover a specified value.
    """

    _option_registry = ['ref']

    @property
    def resolved(self):
        if 'resolved' not in self.state:
            return False
        else:
            return self.state['resolved']

    @resolved.setter
    def resolved(self, value):
        if isinstance(value, bool):
            self.state['resolved'] = value
        else:
            raise TypeError('{0} is not boolean'.format(value))

    def is_resolved(self):
        return self.resolved

    def set_up_search_path(self, name):
        fns = [os.path.join(os.getcwd(), name)]
        try:
            fns.append(os.path.join(self.conf.paths.includes, name))
        except:
            pass

        fns.extend([os.path.join(d[0], name) for d in os.walk(os.getcwd())])

        return fns

    @property
    def file(self):
        return self.state['file']

    @file.setter
    def file(self, value):
        fns = self.set_up_search_path(value)

        for fn in fns:
            if os.path.isfile(fn):
                self.state['file'] = value
                break

        if 'file' not in self.state:
            raise TypeError('file named {0} does not exist'.format(value))


class InheritableContentBase(RecursiveConfigurationBase):
    """
    Base data object that represents a single unit of content. Typically
    sub-classed.
    """

    _option_registry = ['pre', 'post', 'final', 'ref', 'content', 'edition']
    _reference_type = InheritanceReference

    def get_default_replacement(self):
        return {}

    @property
    def replacement(self):
        if 'replacement' not in self.state:
            self.state['replacement'] = self.get_default_replacement()

        return self.state['replacement']

    @replacement.setter
    def replacement(self, value):
        if 'replacement' in self.state:
            if value in ({}, None):
                return
            base = self.state['replacement']
        else:
            base = self.get_default_replacement()

        if isinstance(value, dict):
            base.update(value)
            self.state['replacement'] = base
        elif isinstance(value, collections.Iterable):
            for item in value:
                if len(item) != 2:
                    raise TypeError

            addition = dict(value)
            base.update(addition)
            self.state['replacement'] = base
        else:
            raise TypeError

    @property
    def title(self):
        return self.state['title']

    @title.setter
    def title(self, value):
        if isinstance(value, basestring):
            self.state['title'] = TitleData({'text': value})
        elif isinstance(value, TitleData):
            self.state['title'] = value
        elif isinstance(value, dict):
            self.state['title'] = TitleData(value)
        else:
            raise TypeError

    @property
    def source(self):
        if 'source' in self.state:
            return self.state['source']
        else:
            return None

    @source.setter
    def source(self, value):
        self.state['source'] = self._reference_type(value, self.conf)

    inherit = source

    def is_resolved(self):
        if self.source is None:
            return True
        else:
            try:
                return self.source.resolved
            except AttributeError:
                return False

    def _is_resolveable(self, data):
        try:
            if self.source.file in data.cache:
                return True
            else:
                for key in data.cache.keys():
                    if key.endswith(self.source.file):
                        return True

                return False
        except AttributeError:
            logger.warning(str(data) + ' is not resolvable')
            return False

    def has_field(self, name):
        if name in self and getattr(self, name, None) is not None:
            return True
        else:
            return False

    def resolve(self, data):
        if self.is_resolved() is True:
            return True

        if self._is_resolveable(data):
            try:
                base = copy.deepcopy(data.fetch(self.source.file, self.source.ref))
                base.resolve(data)

                needs_replacement = self.replacement != base.replacement

                if needs_replacement:
                    replacement = copy.deepcopy(base.replacement)
                    replacement.update(self.replacement)

                base.state.update(self.state)
                self.state.update(base.state)

                if needs_replacement:
                    self.replacement = replacement

                self.source.resolved = True

                return True
            except InheritableContentError as e:
                logger.error(e)

        if not self.is_resolved():
            m = 'cannot find {0} and ref "{1}" do not exist'.format(self.source.file,
                                                                    self.source.ref)
            logger.error(m)
            raise InheritableContentError(m)

    def render(self):
        if self.replacement:
            attempts = list(range(10))

            for key in self.state.keys():
                if isinstance(self.state[key], collections.Iterable):
                    should_resplit = None

                    # code blocks are stored internally as lists to preserve
                    # formatting, so to preserve formatting and compatibility
                    # with existing examples, we join them up, do the formatting
                    # and then split it back up.
                    if isinstance(self.state[key], list):
                        for it in self.state[key]:
                            if not isinstance(it, basestring):
                                should_resplit = False
                                break

                        # do the splitting, but if any of the elements weren't
                        # strings then we can't render the string, so it makes
                        # sense to go on to the next key in the dict.
                        if should_resplit is None:
                            should_resplit = True
                            self.state[key] = '\n'.join(self.state[key])
                        elif should_resplit is False:
                            continue

                    # we do this in a loop to in case there are replacements in
                    # replacements. 10 seems like a good number.
                    for i in attempts:
                        if '{{' not in self.state[key]:
                            break

                        template = jinja2.Template(self.state[key])
                        self.state[key] = template.render(**self.replacement)
                        if '{{' not in self.state[key]:
                            break

                    if should_resplit is True:
                        self.state[key] = self.state[key].split('\n')
                elif isinstance(self.state[key], InheritableContentBase):
                    if len(self.state[key].replacement) == 0:
                        self.state[key].replacement = self.replacement

                    self.state[key].render()

                if isinstance(self.state[key], basestring) and "{{" in self.state[key]:
                    logger.error("unable to resolve all tokens in content"
                                 " '{0}' key '{1}'.".format(self.ref, key))


class DataContentBase(RecursiveConfigurationBase):
    """
    Represents a group of units ingested from a single file.
    """

    content_class = InheritableContentBase
    edition_check = staticmethod(lambda x, y: True)

    def __init__(self, src, data, conf):
        self._state = {'content': {}}
        self._content = self._state['content']
        self._conf = None
        self._ordering = []
        self._reordered = False
        self.conf = conf
        self.data = data
        self.ingest(src)

    def __contains__(self, value):
        if value in self.content:
            return True
        else:
            return False

    @property
    def ordering(self):
        return self._ordering

    @property
    def content(self):
        return self.state['content']

    @content.setter
    def content(self, value):
        logger.warning('cannot set content record directly')

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if isinstance(value, DataCache):
            self._data = value
        else:
            logger.warning('cannot use invalid data cache instance.')

    def ingest(self, src):
        if not isinstance(src, list) and os.path.isfile(src):
            with open(src, 'r') as f:
                src = [doc for doc in yaml.safe_load_all(f)]

        for doc in src:
            if doc is None:
                logger.error("content documents cannot be null. check document separators")

            if self.edition_check(doc, self.conf) is False:
                continue
            try:
                content = self.add(doc)
                self._ordering.append(content.ref)
                self._reordered = False
            except RuntimeError:
                pass
            except Exception as e:
                logger.error('could not inherit, because: ' + str(e))
                logger.info(doc)
                raise e

    def add(self, doc):
        if 'ref' in doc:
            ref = doc['ref']
        else:
            if 'source' in doc:
                ref = doc['source']['ref']
            elif 'inherit' in doc:
                ref = doc['inherit']['ref']
            else:
                ref = None

            if ref is not None:
                doc['ref'] = ref

        if ref not in self.content:
            if isinstance(doc, self.content_class):
                content = doc
            else:
                try:
                    content = self.content_class(doc)
                except TypeError:
                    content = self.content_class(doc, self.conf)

            if ref is not None:
                content.ref = ref

            self.content[content.ref] = content

            if not content.is_resolved():
                content.resolve(self.data)

            return content
        else:
            m = 'content named "{0}" already exists'.format(doc['ref'])
            logger.error(m)
            logger.warning(doc)
            raise InheritableContentError(m)

    def fetch(self, ref):
        if ref in self.content:
            content = self.content[ref]

            if not content.is_resolved():
                content.resolve(self)

            return content
        else:
            m = 'content with ref "{0}" not found'.format(ref)
            logger.error(m)
            raise InheritableContentError(m)

    def is_resolved(self):
        unresolved = [1 for exmp in self.content.values()
                      if not exmp.is_resolved()]

        if sum(unresolved) > 0:
            return False
        else:
            return True

    def resolve(self):
        """Resolves all inheritance."""

        for exmp in self.content.values():
            exmp.resolve(self.data)

    def ordered_content(self):
        """
        Generator of content items in order. If items have a ``number`` member,
        then we sort in this order. Otherwise, we return in the order that they
        were specified in the file.
        """
        def sorter(ref_a, ref_b):
            # would be cmp() in python 2, but this is py3 compatible
            a_num = self.fetch(ref_a).number
            b_num = self.fetch(ref_b).number

            return (a_num > b_num) - (a_num < b_num)

        if self._reordered is False:
            for content in self.content.values():
                if 'number' not in content:
                    self._reordered = True
                    break

        if self._reordered is False:
            self._ordering.sort(cmp=sorter)
            self._reordered = True

        for ref in self._ordering:
            yield self.fetch(ref)


class DataCache(RecursiveConfigurationBase):
    """
    Represents a group of related files that hold similar kinds of structured
    data. Often subclassed.
    """

    content_class = DataContentBase
    content_type = None

    def __init__(self, files, conf):
        self._cache = {}
        self._conf = conf
        self.ingest(files)

    def __len__(self):
        return len(self._cache)

    def __contains__(self, key):
        return key in self.cache

    def set_up_search_path(self, name):
        fns = [os.path.join(os.getcwd(), name)]
        try:
            fns.append(os.path.join(self.conf.paths.includes, name))
        except:
            pass

        fns.extend([os.path.join(d[0], name) for d in os.walk(os.getcwd())])

        return fns

    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, value):
        logger.warning('cannot set cache record directly')

    def _clear_cache(self, fn):
        self.cache[fn] = []

    def ingest(self, files):
        setup = [self._clear_cache(fn)
                 for fn in files
                 if fn not in self.cache]

        logger.debug('setup cache for {0} files'.format(len(setup)))

        for fn in files:
            self.add_file(fn)

    def add_file(self, fn):
        if fn not in self.cache or self.cache[fn] == []:
            with open(fn, 'r') as f:
                data = [doc for doc in yaml.safe_load_all(f)]

            self.cache[fn] = self.content_class(data, self, self.conf)
        else:
            logger.debug('populated file {0} exists in the cache'.format(fn))

    def fetch(self, fn, ref):
        if fn in self.cache:
            if self.cache[fn] == []:
                self.add_file(fn)
            return self.cache[fn].fetch(ref)
        else:
            logger.error('file "{0}" is not included.'.format(fn))
            if os.path.isfile(fn):
                self.add_file(fn)
                return self.cache[fn].fetch(ref)
            else:
                logger.warning("file ({0}) doesn't exist, crawling directory".format(fn))
                fns = self.set_up_search_path(fn)

                for filen in fns:
                    if os.path.isfile(filen):
                        self.add_file(filen)
                        break

                if filen in self.cache:
                    return self.cache[filen].fetch(ref)
                else:
                    raise InheritableContentError('cannot resolve: {0} {1}'.format(fn, ref))

    def file_iter(self):
        for fn in self.cache:
            yield fn, self.cache[fn]

    def content_iter(self):
        for fn in self.cache:
            for data in self.cache[fn].content.values():
                if data.ref.startswith('_'):
                    continue
                else:
                    yield fn, data


class TitleData(ConfigurationBase):
    _option_registry = ['text']

    level_characters = {"=": 1,
                        "-": 2,
                        "~": 3,
                        "`": 4,
                        "^": 5,
                        "'": 6}

    @property
    def character(self):
        return self.level

    @character.setter
    def character(self, value):
        self.level = self.level_characters[value]

    @property
    def level(self):
        if 'level' not in self.state:
            return 3
        else:
            return self.state['level']

    @level.setter
    def level(self, value):
        if isinstance(value, int):
            self.state['level'] = value
        else:
            raise TypeError



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Responsible for converting the Sphinx configuration stored by projects and used
by giza, into a ``sphinx-build`` invocation. See
:func:`giza.content.sphinx.run_sphinx` for the core of this operation.

:mod:`giza.content.sphinx` also contains:

- the error message processing, which normalizes references to paths,
  deduplicates log messages given multiple builds or multi-process Sphinx
  operation, and removes non-actionable messages. See
  :func:`giza.content.sphinx.output_sphinx_stream()`,
  :func:`giza.content.sphinx.is_msg_worth()` and
  :func:`giza.content.sphinx.path_normalization()`.

- the orchestration of all post-processing. See
  :func:`giza.content.sphinx.finalize_sphinx_build()`.
"""

import collections
import logging
import os.path
import pkg_resources
import re
import numbers
import subprocess
import shlex

from giza.libgiza.task import Task
from giza.tools.files import safe_create_directory, expand_tree
from giza.tools.timing import Timer

logger = logging.getLogger('giza.content.sphinx')

# Config Resolution


def is_parallel_sphinx(version):
    return version >= '1.2'


def get_tags(target, sconf):
    if 'tags' in sconf:
        ret = set(sconf.tags)
    else:
        ret = set()

    ret.add(target)
    ret.add(target.split('-')[0])

    if target.startswith('html') or target.startswith('dirhtml'):
        ret.add('website')
    else:
        ret.add('print')

    if 'edition' in sconf:
        ret.add(sconf.edition)

    return ' '.join([' '.join(['-t', i])
                     for i in ret
                     if i is not None])


def get_sphinx_args(sconf, conf):
    o = []

    o.append(get_tags(sconf.builder, sconf))
    o.append('-q')

    o.append('-b {0}'.format(sconf.builder))
    n_workers = str(min(conf.runstate.pool_size, 4))

    if is_parallel_sphinx(pkg_resources.get_distribution("sphinx").version):
        if 'serial_sphinx' in conf.runstate:
            m = 'running with serial sphinx processes ({0}.{1}.{2}.{3})'
            logger.debug(m.format(sconf.builder, conf.project.name,
                                 conf.project.edition, conf.git.branches.current))
            if conf.runstate.serial_sphinx == "publish":
                if ((len(conf.runstate.builder) >= 1 or 'publish' in conf.runstate.builder) or
                        len(conf.runstate.languages_to_build) >= 1 or
                        len(conf.runstate.editions_to_build) >= 1):
                    pass
                else:
                    o.append(' '.join(['-j', n_workers]))
            elif conf.runstate.serial_sphinx is False:
                logger.debug('running with parallelized sphinx processes')
                o.append(' '.join(['-j', n_workers]))
            elif (isinstance(conf.runstate.serial_sphinx, numbers.Number) and
                  conf.runstate.serial_sphinx > 1):
                logger.debug('running with parallelized sphinx processes')
                o.append(' '.join(['-j', n_workers]))
            else:
                pass
        elif len(conf.runstate.builder) >= conf.runstate.pool_size:
            logger.debug('running with serial sphinx processes')
            pass
        else:
            logger.debug('running with parallelized sphinx processes')
            o.append(' '.join(['-j', n_workers]))

    o.append(' '.join(['-c', conf.paths.projectroot]))

    if 'language' in sconf and sconf.language is not None:
        o.append("-D language='{0}'".format(sconf.language))

    return ' '.join(o)

# Output Management


def output_sphinx_stream(out, conf):
    full_path = os.path.join(conf.paths.projectroot, conf.paths.branch_output)

    regx = r'(.*):[0-9]+: WARNING: duplicate object description of ".*", other instance in (.*)'
    regx = re.compile(regx)

    printable = []
    for idx, l in enumerate(out):
        if is_msg_worthy(l) is False:
            printable.append(None)
            continue

        f1 = regx.match(l)
        if f1 is not None:
            g = f1.groups()

            if g[1].endswith(g[0]):
                printable.append(None)
                continue

        l = path_normalization(l, full_path, conf)

        if l.startswith('InputError: [Errno 2] No such file or directory'):
            try:
                l = path_normalization(l.split(' ')[-1].strip()[1:-2], full_path, conf)
                printable[idx - 1] += ' ' + l
                l = None
            except IndexError:
                logger.error("error processing log: {0}".format(l))
                continue
        elif l.startswith('source/includes/generated/overview.rst'):
            continue
        elif l.startswith('source/meta/includes.txt'):
            continue

        printable.append(l)

    printable = stable_deduplicate(printable)

    m = 'sphinx builder has {0} lines of output, processed from {1}'
    logger.info(m.format(len(printable), len(out)))
    print_build_messages(printable)


def stable_deduplicate(lines):
    # this should probably just use OrderedSet() in the future

    mapping = collections.OrderedDict()

    for idx, ln in enumerate(lines):
        mapping[ln] = idx

    return list(mapping.keys())


def print_build_messages(messages):
    for l in (l for l in messages if l is not None):
        if 'WARNING: ' in l:
            logger.warn(l, extra={'lean': True})
        elif 'ERROR: ' in l or 'SEVERE: ' in l:
            logger.error(l, extra={'lean': True})
        else:
            logger.info(l, extra={'lean': True})


def path_normalization(l, full_path, conf):
    if l.startswith('..'):
        l = os.path.sep.join([el for el in l.split(os.path.sep)
                              if el != '..'])

    if l.startswith(conf.paths.branch_output):
        l = l[len(conf.paths.branch_output) + 1:]
    elif l.startswith(full_path):
        l = l[len(full_path) + 1:]

    if l.startswith('source'):
        l = os.path.sep.join(['source', l.split(os.path.sep, 1)[1]])

    return l


def is_msg_worthy(l):
    if l.startswith('WARNING: unknown mimetype'):
        return False
    elif len(l) == 0:
        return False
    elif l.startswith('WARNING: search index'):
        return False
    elif 'WARNING: Could not lex literal_block as' in l:
        return False
    elif 'extension is not safe for parallel reading, doing serial read' in l:
        return False
    elif l.endswith('source/reference/sharding-commands.txt'):
        return False
    elif l.endswith('Duplicate ID: "cmdoption-h".'):
        return False
    elif l.endswith('"/opt args" or "+opt args"'):
        return False
    elif l.endswith('"--opt args" or "/opt args"'):
        return False
    elif 'nonlocal image URI found' in l:
        return False
    else:
        return True


# Builder Operation


def run_sphinx(builder, sconf, conf):
    if safe_create_directory(sconf.fq_build_output):
        m = 'created directory "{1}" for sphinx builder {0}'
        logger.debug(m.format(builder, sconf.fq_build_output))

    if 'language' in sconf and sconf.language is not None:
        cmd_str = 'sphinx-intl build --language=' + sconf.language
        try:
            subprocess.check_call(shlex.split(cmd_str))
            logger.info('compiled all PO files for translated build.')
        except subprocess.CalledProcessError as e:
            logger.error('sphinx-intl encountered error: ' + str(e.returncode))
            logger.info(cmd_str)

    cmd = 'sphinx-build {0} -d {1}/doctrees-{2} {3} {4}'  # per-builder-doctree

    sphinx_cmd = cmd.format(get_sphinx_args(sconf, conf),
                            os.path.join(conf.paths.projectroot, conf.paths.branch_output),
                            sconf.build_output,
                            os.path.join(conf.paths.projectroot, conf.paths.branch_source),
                            sconf.fq_build_output)

    logger.info('Starting sphinx build: ' + sphinx_cmd)
    m = "running sphinx build for: {0}, {1}, {2}"

    with Timer(m.format(builder, sconf.language, sconf.edition)):
        try:
            output = subprocess.check_output(shlex.split(sphinx_cmd), stderr=subprocess.STDOUT)
            return_code = 0
        except subprocess.CalledProcessError as e:
            output = e.output
            return_code = e.returncode
            logger.info(sphinx_cmd)
    try:
        os.utime(sconf.fq_build_output, None)
    except:
        pass

    m = 'completed {0} sphinx build for {1}.{2}.{3} ({4})'

    logger.info(m.format(builder, conf.project.name, conf.project.edition,
                         conf.git.branches.current, return_code))

    return return_code, output

# Application Logic


def sphinx_tasks(sconf, conf):
    # Projects that use the append functionality in extracts or similar content
    # generators will rebuild this task every time.

    deps = [os.path.join(conf.paths.projectroot, 'conf.py')]
    deps.extend(conf.system.files.get_configs('sphinx_local'))
    deps.extend(expand_tree(os.path.join(conf.paths.projectroot, conf.paths.branch_source), 'txt'))

    return Task(job=run_sphinx,
                args=(sconf.builder, sconf, conf),
                target=os.path.join(conf.paths.projectroot,
                                    conf.paths.branch_output,
                                    sconf.builder),
                dependency=deps,
                description='building {0} with sphinx'.format(sconf.builder))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Entry points that run the content generation tasks. Typically use these entry
points for testing and for prepopulating build directories. These functions do
not run in common practice.
"""

import logging

import argh

from giza.libgiza.app import BuildApp

from giza.config.helper import fetch_config, get_restricted_builder_jobs

from giza.content.assets import assets_tasks, assets_clean
from giza.content.images.tasks import image_tasks, image_clean
from giza.content.intersphinx import intersphinx_tasks, intersphinx_clean
from giza.content.robots import robots_txt_tasks
from giza.content.redirects import make_redirect, redirect_tasks
from giza.content.migrations import migration_tasks, migration_clean

from giza.content.tocs.tasks import toc_tasks
from giza.content.apiargs.tasks import apiarg_tasks
from giza.content.examples.tasks import example_tasks
from giza.content.steps.tasks import step_tasks, step_clean
from giza.content.options.tasks import option_tasks, option_clean
from giza.content.release.tasks import release_tasks, release_clean
from giza.content.glossary.tasks import glossary_tasks, glossary_clean
from giza.content.changelog.tasks import changelog_tasks

from giza.operations.sphinx_cmds import sphinx_content_preperation, sphinx_builder_tasks

logger = logging.getLogger('giza.operations.generate')


@argh.arg('--edition', '-e')
@argh.expects_obj
def toc(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        app.extend_queue(toc_tasks(c))


@argh.arg('--edition', '-e')
@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def steps(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        if c.runstate.clean_generated is True:
            app.extend_queue(step_clean(c))
        else:
            app.extend_queue(step_tasks(c))


@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def options(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:

        if c.runstate.clean_generated is True:
            app.extend_queue(option_clean(c))
        else:
            app.extend_queue(option_tasks(c))


@argh.expects_obj
def api(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        app.extend_queue(apiarg_tasks(c))


@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def assets(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        if c.runstate.clean_generated is True:
            app.extend_queue(assets_clean(c))
        else:
            app.extend_queue(assets_tasks(c))


@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*', dest='languages_to_build')
@argh.arg('--builder', '-b', nargs='*', default='html')
@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def images(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:

        if c.runstate.clean_generated is True:
            app.extend_queue(image_clean(c))
        else:
            for (_, (bconf, sconf)) in get_restricted_builder_jobs(c):
                app.extend_queue(image_tasks(bconf, sconf))


@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def intersphinx(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        if c.runstate.clean_generated is True:
            app.extend_queue(intersphinx_clean(c))
        else:
            app.extend_queue(intersphinx_tasks(c))


@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.aliases('primer', 'migrations')
@argh.expects_obj
def migration(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        if c.runstate.clean_generated is True:
            app.extend_queue(migration_clean(c))
        else:
            app.extend_queue(migration_tasks(c))


@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def release(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        if c.runstate.clean_generated is True:
            app.extend_queue(release_clean(c))
        else:
            app.extend_queue(release_tasks(c))


@argh.expects_obj
def examples(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        app.extend_queue(example_tasks(c))


@argh.expects_obj
def changelogs(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        app.extend_queue(changelog_tasks(c))


@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def glossary(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        if c.runstate.clean_generated is True:
            app.extend_queue(glossary_clean(c))
        else:
            app.extend_queue(glossary_tasks(c))


@argh.arg('--edition', '-e')
@argh.expects_obj
def robots(args):
    c = fetch_config(args)

    with BuildApp.new(pool_type=c.runstate.runner,
                      pool_size=c.runstate.pool_size,
                      force=c.runstate.force).context() as app:
        app.pool = 'serial'
        app.extend_queue(robots_txt_tasks(c))


@argh.arg('--edition', '-e')
@argh.arg('--print', '-p', default=False, action='store_true', dest='dry_run')
@argh.expects_obj
def redirects(args):
    c = fetch_config(args)

    if args.dry_run is True:
        print(''.join(make_redirect(c)))
    else:
        with BuildApp.new(pool_type=c.runstate.runner,
                          pool_size=c.runstate.pool_size,
                          force=c.runstate.force).context() as app:
            app.extend_queue(redirect_tasks(c))


@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*', dest='languages_to_build')
@argh.expects_obj
def source(args):
    args.builder = 'html'
    conf = fetch_config(args)

    with BuildApp.new(pool_type=conf.runstate.runner,
                      pool_size=conf.runstate.pool_size,
                      force=conf.runstate.force).context() as app:
        sphinx_content_preperation(app, conf)


@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*', dest='languages_to_build')
@argh.arg('--builder', '-b', nargs='*', default='html')
@argh.expects_obj
def sphinx(args):
    if args.runner == 'serial':
        args.serial_sphinx = True

    conf = fetch_config(args)
    logger.warning('not for production use: this expects that content generation is complete.')

    app = BuildApp.new(pool_type=conf.runstate.runner,
                       pool_size=conf.runstate.pool_size,
                       force=conf.runstate.force)

    r = sphinx_builder_tasks(app, conf)

    raise SystemExit(r)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`~giza.content.assets` makes it possible to configure embedded `git`
repositories within a ``giza`` configured project. These repositories may either
be updated to the latest revision upon every build or are pinned to a specific
revision in the configuration file.

Assets specifications are in the top level of the build configuration accessible
via the ``assets`` field in a :class:`~giza.config.main.Configuration` instance,
which holds a list of repository definitions, which resembles the following: ::

   {
     "branch": <string>,
     "path": <path>,
     "repository": <url>
   }


.. describe:: assets.branch

   The name of a remote branch in the repository.

.. describe:: assets.path

   A local path for the cloned repository. Relative to the top of the repository.

.. describe:: assets.repository

   A git-compatible URL for the source repository.

A future version of the assets system may allow users to specify a specific
revision rather than a branch so that builds for specific branches are more
stable over time.
"""

import logging
import os
import shutil
import subprocess

import giza.libgiza.task
import giza.libgiza.git

import giza.tools.files

logger = logging.getLogger('giza.content.assets')


def assets_setup(path, branch, repo, commit=None):
    """
    Worker function that clones a repository if one doesn't exist and pulls
    the repository otherwise.
    """
    # TODO: In the future this should be able to pin the repository to a
    #       specific hash.

    if os.path.exists(path):
        g = giza.libgiza.git.GitRepo(path)

        if commit is None:
            try:
                g.pull(branch=branch)
            except giza.libgiza.git.GitError as error:
                logger.error('failed to pull %s repository', path)
                logger.debug(error)
                return

            logger.info('updated %s repository', path)
            return
        elif g.sha() == commit or g.sha().startswith(commit):
            logger.info('repository %s is currently at (%s)', path, commit)
        else:
            g.checkout(commit)
            logger.info('update %s repository to (%s)', path, commit)
    else:
        base, name = os.path.split(path)
        giza.tools.files.safe_create_directory(base)

        g = giza.libgiza.git.GitRepo(base)
        g.clone(repo, repo_path=path, branch=branch)
        logger.info('cloned %s branch from repo %s', branch, repo)

        if commit is not None and (g.sha() == commit or g.sha().startswith(commit)):
            g.checkout(commit)
            logger.info('repository %s is currently at (%s)', path, commit)


def assets_tasks(conf):
    """Add tasks to an app to create/update the assets."""

    tasks = []
    generate_tasks = []
    if conf.assets is not None:
        giza.tools.files.safe_create_directory(conf.paths.projectroot)
        for asset in conf.assets:
            path = os.path.join(conf.paths.projectroot, asset.path)

            logger.debug('adding asset resolution job for %s', path)

            args = {'path': path,
                    'branch': asset.branch,
                    'repo': asset.repository}

            if 'commit' in asset:
                args['commit'] = asset.commit

            description = "setup assets for: {0} in {1}".format(asset.repository, path)
            tasks.append(giza.libgiza.task.Task(job=assets_setup,
                                                args=args,
                                                target=path,
                                                dependency=None,
                                                description=description))

            # If you specify a list of "generate" items, giza will call ``giza
            # generate`` to build content after updating the
            # repository. Deprecated, and largely unused.
            if 'generate' in asset:
                for content_type in asset.generate:
                    description = 'generating objects in {0}'.format(path)
                    args = dict(cwd=path, args=['giza', 'generate', content_type])
                    generate_tasks.append(giza.libgiza.task.Task(job=subprocess.call,
                                                                 target=path,
                                                                 dependency=None,
                                                                 args=args,
                                                                 description=description))

    if len(generate_tasks) > 0:
        tasks.append(generate_tasks)

    return tasks


def assets_clean(conf):
    """Adds tasks to remove all asset repositories."""

    tasks = []

    if conf.assets is not None:
        for asset in conf.assets:
            path = os.path.join(conf.paths.projectroot, asset.path)
            logger.debug('adding asset cleanup %s', path)

            t = giza.libgiza.task.Task(job=shutil.rmtree,
                                       args=[path],
                                       target=path,
                                       dependency=None,
                                       description='cleaning up asset: ' + path)

            tasks.append(t)

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.tools.transformation import append_to_file, prepend_to_file
from giza.content.extract.inheritance import ExtractDataCache
from giza.content.extract.views import render_extracts, get_include_statement
from giza.config.content import new_content_type
from giza.libgiza.task import Task

logger = logging.getLogger('giza.content.extract.tasks')


def register_extracts(conf):
    content_dfn = new_content_type(name='extracts',
                                   task_generator=extract_tasks,
                                   conf=conf)

    conf.system.content.add(name='extracts', definition=content_dfn)


def write_extract_file(extract, fn):
    content = render_extracts(extract)
    content.write(fn)
    logger.debug('wrote extract file: ' + fn)


def extract_tasks(conf):
    extracts = ExtractDataCache(conf.system.content.extracts.sources, conf)
    extracts.create_output_dir()

    tasks = []
    for dep_fn, extract in extracts.content_iter():
        t = Task(job=write_extract_file,
                 args=(extract, extract.target),
                 description="generating extract file: " + extract.target,
                 target=extract.target,
                 dependency=dep_fn)
        tasks.append(t)

        include_statement = get_include_statement(extract.target_project_path)

        for verb, adjc, files in [(prepend_to_file, 'prepend', extract.prepend),
                                  (append_to_file, 'append', extract.append)]:
            # have to run appends and prepends always, because the rsync that
            # populates build/<branch>/source should and does overwrite these
            # files on every source generation step. None in the dep list does this.
            for fn in files:
                msg = "{} extract include for '{}' to '{}'".format(adjc, extract.target, fn)
                t = Task(job=verb,
                         args=(fn, include_statement),
                         target=fn,
                         dependency=[None, dep_fn],
                         description=msg)
                tasks.append(t)

    logger.debug('added tasks for {0} extract generation tasks'.format(len(tasks)))

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
import random
from unittest import TestCase

from giza.libgiza.app import BuildApp
from giza.libgiza.pool import ThreadPool, ProcessPool, SerialPool
from giza.libgiza.task import Task
from giza.config.main import Configuration
from giza.config.runtime import RuntimeStateConfig


class CommonAppSuite(object):
    def test_add_make_test_default(self):
        self.assertEqual(self.app.queue, [])
        self.app.add()
        self.assertTrue(len(self.app.queue) == 1)
        self.assertTrue(isinstance(self.app.queue[0], Task))

    def test_add_make_test_task(self):
        self.assertEqual(self.app.queue, [])
        self.app.add('task')
        self.assertTrue(len(self.app.queue) == 1)
        self.assertTrue(isinstance(self.app.queue[0], Task))

    def test_add_make_test_app(self):
        self.assertEqual(self.app.queue, [])
        self.app.add('app')
        self.assertTrue(len(self.app.queue) == 1)
        self.assertIsInstance(self.app.queue[0], BuildApp)

    def test_add_existing_task_object(self):
        self.assertEqual(self.app.queue, [])
        t = Task()
        self.app.add(t)
        self.assertIs(t, self.app.queue[0])
        self.assertIsNot(t, Task())
        self.assertIsNot(Task(), self.app.queue[0])

    def test_pool_setter_default(self):
        self.assertIsNone(self.app.worker_pool)
        pool_type = self.app.default_pool

        if pool_type == 'lazy':
            pool_type = random.choice(['thread', 'serial'])

        self.app.pool = pool_type

        self.assertIsNone(self.app.worker_pool)
        self.app.create_pool()
        self.assertIsNotNone(self.app.worker_pool)
        self.assertIsInstance(self.app.pool, self.app.pool_mapping[pool_type])

    def test_pool_setter_process(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.pool = 'process'
        self.assertIsNone(self.app.worker_pool)
        self.app.create_pool()
        self.assertIsNotNone(self.app.worker_pool)
        self.assertIsInstance(self.app.pool, ProcessPool)
        self.assertTrue(self.app.has_active_pool())

    def test_pool_setter_thread(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.pool = 'thread'
        self.assertIsNone(self.app.worker_pool)
        self.app.create_pool()
        self.assertIsNotNone(self.app.worker_pool)
        self.assertIsInstance(self.app.pool, ThreadPool)
        self.assertTrue(self.app.has_active_pool())

    def test_pool_setter_serial(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.pool = 'serial'
        self.assertIsNone(self.app.worker_pool)
        self.app.create_pool()
        self.assertIsNotNone(self.app.worker_pool)
        self.assertIsInstance(self.app.pool, SerialPool)
        self.assertTrue(self.app.has_active_pool())

    def test_pool_setter_process_by_ref(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.pool = ProcessPool
        self.assertIsNotNone(self.app.worker_pool)
        self.assertIsInstance(self.app.pool, ProcessPool)
        self.assertTrue(self.app.has_active_pool())

    def test_pool_setter_thread_by_ref(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.pool = ThreadPool
        self.assertIsNotNone(self.app.worker_pool)
        self.assertIsInstance(self.app.pool, ThreadPool)
        self.assertTrue(self.app.has_active_pool())

    def test_pool_setter_serial_by_ref(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.pool = SerialPool
        self.assertIsNotNone(self.app.worker_pool)
        self.assertIsInstance(self.app.pool, SerialPool)
        self.assertTrue(self.app.has_active_pool())

    def test_pool_setter_invalid_input(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.default_pool = 'serial'

        self.app.pool = 1
        self.app.create_pool()
        self.assertIn(type(self.app.pool), self.app.pool_mapping.values())
        self.assertTrue(self.app.has_active_pool())

    def test_pool_closer(self):
        self.assertIsNone(self.app.worker_pool)
        self.app.pool = 'thread'
        self.assertIsNone(self.app.worker_pool)
        self.app.create_pool()
        self.assertIsInstance(self.app.pool, ThreadPool)
        self.assertTrue(self.app.has_active_pool())
        self.app.close_pool()
        self.assertIsNone(self.app.worker_pool)
        self.assertFalse(self.app.has_active_pool())

    def test_add_invalid_object(self):
        with self.assertRaises(TypeError):
            self.app.add(1)

    def test_single_runner_task(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        t = Task()
        t.job = sum
        t.description = 'test task'
        t.args = [[1, 2], 0]

        self.app.add(t)
        self.app.run(t)
        self.assertEqual(self.app.results[0], 3)

    def test_single_runner_task_integrated(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        t = self.app.add('task')

        t.job = sum
        t.args = [[1, 2], 0]
        t.description = 'test task'

        self.app.add(t)
        self.app.run()
        self.assertEqual(self.app.results[0], 3)

    def test_single_runner_app_integrated(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        app = self.app.add('app')

        t = app.add('task')
        t.job = sum
        t.description = 'test task'
        t.args = [[1, 2], 0]

        self.app.run()
        self.assertEqual(self.app.results[0], 3)

    def test_results_ordering(self):
        expected_results = [12, 13, 14, 15, 7, 17, 18, 10, 20, 12]

        self.assertIsNone(self.app.pool)
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for inc in range(10):
            t = self.app.add('task')
            t.job = sum
            if inc in (4, 7, 9):
                t.args = [[1, 2, inc], 0]
            else:
                t.args = [[20, 2, inc - 10], 0]
            t.description = 'test task'

        self.app.run()
        self.assertEqual(self.app.results, expected_results)

    def test_single_runner_app_integrated_with_many_subtasks(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        app = self.app.add('app')

        for _ in range(10):
            t = app.add('task')
            t.job = sum
            t.args = [[1, 2], 0]
            t.description = 'test task'

        self.app.run()
        self.assertEqual(len(self.app.results), 10)
        self.assertEqual(self.app.results[0], 3)
        self.assertEqual(sum(self.app.results), 30)

    def test_has_apps_predicate_single(self):
        self.assertEqual(self.app.queue, [])

        self.app.queue.append(None)
        self.assertFalse(self.app.queue_has_apps)

    def test_has_apps_predicate_empty(self):
        self.assertEqual(self.app.queue, [])
        self.assertFalse(self.app.queue_has_apps)

    def test_has_apps_predicate_all_tasks(self):
        self.assertEqual(self.app.queue, [])

        for _ in range(10):
            self.app.add('task')

        self.assertEqual(len(self.app.queue), 10)
        self.assertFalse(self.app.queue_has_apps)

    def test_has_apps_with_member_tasks_predicate_all_apps(self):
        self.assertEqual(self.app.queue, [])

        for _ in range(10):
            a = self.app.add('app')
            self.assertIsInstance(a, BuildApp)
            t = a.add('task')
            t.job = sum
            t.args = (1, 2)

        self.assertEqual(len(self.app.queue), 10)
        self.assertTrue(self.app.queue_has_apps)

    def test_has_apps_without_member_tasks_predicate_all_apps(self):
        self.assertEqual(self.app.queue, [])

        for _ in range(10):
            self.app.add('app')

        self.assertEqual(len(self.app.queue), 10)
        self.assertTrue(self.app.queue_has_apps)

    def test_has_apps_predicate_mixed(self):
        self.assertEqual(self.app.queue, [])

        for _ in range(10):
            self.app.add('task')

        for _ in range(10):
            a = self.app.add('app')
            self.assertIsInstance(a, BuildApp)
            t = a.add('task')
            t.job = sum
            t.args = (1, 2)

        self.assertTrue(self.app.queue_has_apps)

    def test_running_mixed_queue_all_apps_integrated(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(10):
            app = self.app.add('app')
            for _ in range(10):
                t = app.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

        self.app.run()

        self.assertEqual(len(self.app.queue), 0)
        self.assertEqual(sum(self.app.results), 300)

    def test_running_mixed_queue_mixed_queue_integrated(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(10):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 2], 0]
            t.description = 'test task'

        for _ in range(10):
            app = self.app.add('app')
            for _ in range(10):
                t = app.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

        self.app.run()

        self.assertEqual(len(self.app.queue), 0)
        self.assertEqual(sum(self.app.results), 330)

    def test_running_mixed_queue_all_apps_direct(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(10):
            app = self.app.add('app')
            for _ in range(10):
                t = app.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

        self.app.create_pool('thread')
        self.app._run_mixed_queue()

        self.assertEqual(sum(self.app.results), 300)
        self.assertEqual(len(self.app.queue), 10)

    def test_running_mixed_queue_mixed_queue_direct(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        self.app.default_pool = 'thread'
        self.app.create_pool()

        for _ in range(10):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 2], 0]
            t.description = 'test task'

        for _ in range(10):
            app = self.app.add('app')
            app.pool = self.app.pool
            for _ in range(10):
                t = app.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

        self.app._run_mixed_queue()

        self.assertEqual(len(self.app.queue), 20)
        self.assertEqual(sum(self.app.results), 330)

    def test_running_tasks_ordering_serial(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(5):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 2], 0]
            t.description = 'test task'

        for _ in range(5):
            t = self.app.add('task')
            t.job = sum
            t.args = [[2, 2], 0]
            t.description = 'test task'

        self.app.create_pool()
        self.app.run()

        self.assertEqual(len(self.app.queue), 0)
        self.assertEqual(self.app.results, [3, 3, 3, 3, 3, 4, 4, 4, 4, 4])

    def test_task_results_ordering_with_apps(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(3):
            app = self.app.add('app')
            for _ in range(5):
                t = app.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

            for _ in range(5):
                t = app.add('task')
                t.job = sum
                t.args = [[2, 2], 0]
                t.description = 'test task'

        self.app.run()

        self.assertEqual(self.app.results,
                         [3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
                          3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
                          3, 3, 3, 3, 3, 4, 4, 4, 4, 4])

    def test_task_results_ordering_varried_with_apps(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        app = self.app.add('app')
        t = app.add('task')
        t.job = sum
        t.args = [[1, 8], 0]
        t.description = 'test task'

        for _ in range(3):
            app = self.app.add('app')
            for _ in range(5):
                t = app.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

            for _ in range(5):
                t = app.add('task')
                t.job = sum
                t.args = [[2, 2], 0]
                t.description = 'test task'

        app = self.app.add('app')
        t = app.add('task')
        t.job = sum
        t.args = [[2, 8], 0]

        for _ in range(5):
            t = app.add('task')
            t.job = sum
            t.args = [[2, 2], 0]
            t.description = 'test task'

        self.app.run()

        expected_results = [9,
                            3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
                            3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
                            3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
                            10, 4, 4, 4, 4, 4]

        self.assertEqual(sorted(self.app.results), sorted(expected_results))
        self.assertEqual(self.app.results, expected_results)

    def test_task_results_lack_of_order(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(5):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 2], 0]
            t.description = 'test task'

        for _ in range(5):
            t = self.app.add('task')
            t.job = sum
            t.args = [[2, 2], 0]
            t.description = 'test task'

        self.app.run()

        # there's a small chance that this could randomly fail without
        # indicating a correctness bug.
        self.assertNotEqual(self.app.results,
                            [3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
                             3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
                             3, 3, 3, 3, 3, 4, 4, 4, 4, 4])

    def test_task_results_task_and_apps0(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(6):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 1], 0]
            t.description = 'test task'

        for _ in range(3):
            app0 = self.app.add('app')
            for _ in range(5):
                t = app0.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 1], 0]
            t.description = 'test task'

            app1 = self.app.add('app')
            for _ in range(5):
                t = app1.add('task')
                t.job = sum
                t.args = [[2, 2], 0]
                t.description = 'test task'

        for _ in range(10):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 1], 0]
            t.description = 'test task'

        self.app.run()

        self.assertEqual(self.app.results,
                         [2, 2, 2, 2, 2, 2,
                          3, 3, 3, 3, 3, 2, 4, 4, 4, 4, 4,
                          3, 3, 3, 3, 3, 2, 4, 4, 4, 4, 4,
                          3, 3, 3, 3, 3, 2, 4, 4, 4, 4, 4,
                          2, 2, 2, 2, 2, 2, 2, 2, 2, 2])

    def test_task_results_task_and_apps1(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        for _ in range(6):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 1], 0]
            t.description = 'test task'

        for _ in range(3):
            app = self.app.add('app')
            for _ in range(5):
                t = app.add('task')
                t.job = sum
                t.args = [[1, 2], 0]
                t.description = 'test task'

            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 1], 0]
            t.description = 'test task'

            for _ in range(5):
                t = app.add('task')
                t.job = sum
                t.args = [[2, 2], 0]
                t.description = 'test task'

        for _ in range(10):
            t = self.app.add('task')
            t.job = sum
            t.args = [[1, 1], 0]
            t.description = 'test task'

        self.app.run()

        self.assertEqual(self.app.results,
                         [2, 2, 2, 2, 2, 2,
                          3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 2,
                          3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 2,
                          3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 2,
                          2, 2, 2, 2, 2, 2, 2, 2, 2, 2
                          ])

    def test_single_runner_app(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        app = BuildApp()
        app.pool_size = 2
        t = app.add('task')
        t.job = sum
        t.args = [[1, 2], 0]
        t.description = 'test task'

        self.app.add(app)
        self.app.run()
        self.assertEqual(self.app.results[0], 3)

    def test_single_runner_app_with_many_subtasks(self):
        self.assertEqual(self.app.queue, [])
        self.assertEqual(self.app.results, [])

        app = BuildApp()
        app.pool_size = 2

        for _ in range(10):
            t = app.add('task')
            t.job = sum
            t.description = 'test task'
            t.args = [[1, 2], 0]

        self.app.add(app)
        self.app.run()
        self.assertEqual(len(self.app.results), 10)
        self.assertEqual(self.app.results[0], 3)
        self.assertEqual(sum(self.app.results), 30)

    def test_add_existing_app_object(self):
        self.assertEqual(self.app.queue, [])
        app = BuildApp()
        app.pool_size = 2
        self.app.add(app)
        self.assertIs(app, self.app.queue[0])
        self.assertIsNot(app, BuildApp())
        self.assertIsNot(BuildApp(), self.app.queue[0])

    def test_pool_setter_existing_pool_thread(self):
        self.assertIsNone(self.app.worker_pool)
        p = ThreadPool(self.c)
        p.pool_size = 2
        self.app.pool = p
        self.assertIs(self.app.pool, p)

    def test_pool_setter_existing_pool_process(self):
        self.assertIsNone(self.app.worker_pool)
        p = ProcessPool(self.c)
        p.pool_size = 2
        self.app.pool = p
        self.assertIs(self.app.pool, p)

    def test_pool_setter_existing_pool_serial(self):
        self.assertIsNone(self.app.worker_pool)
        p = SerialPool(self.c)
        self.app.pool = p
        self.assertIs(self.app.pool, p)

    def test_pool_clenser_removes_empty_apps(self):
        self.assertEqual(len(self.app.queue), 0)

        a = self.app.add('app')
        self.assertIsInstance(a, BuildApp)

        self.assertEqual(len(self.app.queue), 1)
        self.app.clean_queue()
        self.assertEqual(len(self.app.queue), 0)

    def test_finalizers_simple(self):
        t = Task(job=sum,
                 args=((1, 2, 3), 0))

        self.assertEqual(t.finalizers, [])

        t.finalizers = [
            Task(job=sum,
                 args=((4, 5, 6, i), 0))
            for i in range(10)
        ]

        self.app.add(t)
        results = self.app.run()
        self.assertEqual(len(results), 11)
        self.assertEqual(results,
                         [6, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24])

    def test_finalizers_nested(self):
        t = Task(job=sum,
                 args=((1, 2, 3), 0))

        self.assertEqual(t.finalizers, [])

        t.finalizers = [
            Task(job=sum,
                 args=((4, 5, 6, i), 0))
            for i in range(10)
        ]
        self.assertEqual(len(t.finalizers), 10)

        for task in t.finalizers:
            task.finalizers.extend([
                Task(job=sum,
                     args=((4, 5, 6, i), 0))
                for i in range(10)
            ])
            self.assertEqual(len(task.finalizers), 10)

        self.app.add(t)
        results = self.app.run()
        self.assertEqual(len(results), 111)
        self.assertEqual(results[0], 6)
        self.assertEqual(results[1], 15)

        for result in results:
            self.assertIsInstance(result, numbers.Number)
            self.assertTrue(result >= 6)
            self.assertTrue(result <= 24)

    def test_dependency(self):
        self.assertIsNone(self.app.dependency)

        test_value = "string"
        self.app.dependency = test_value
        self.assertIsNotNone(self.app.dependency)
        self.assertIs(test_value, self.app.dependency)

    def test_target(self):
        self.assertIsNone(self.app.target)

        test_value = "string"
        self.app.target = test_value
        self.assertIsNotNone(self.app.target)
        self.assertIs(test_value, self.app.target)

    def test_description(self):
        # if no jobs defined.

        sub_app = self.app.add("app")

        self.assertTrue("member" in sub_app.description)
        self.assertTrue("root level" in self.app.description)

        self.assertEquals(0, len(sub_app.queue))
        self.assertTrue(sub_app.description.endswith(": []"))
        self.assertEquals(1, len(self.app.queue))
        self.assertTrue(self.app.description.endswith(": [{0}]".format(type(self.app))))


class TestBuildAppStandardConfig(CommonAppSuite, TestCase):
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.app = BuildApp(self.c)
        self.app.default_pool = random.choice(['serial', 'thread'])
        self.app.pool_size = 2

    def test_conf_object_consistent_in_task(self):
        self.assertEqual(self.app.queue, [])
        t = self.app.add('task')
        self.assertIs(self.c, t.conf)
        self.assertIs(self.c, self.app.queue[0].conf)

    def test_conf_object_consistent_in_app(self):
        self.assertEqual(self.app.queue, [])
        self.app.add('app')

        self.assertIs(self.c, self.app.conf)
        self.assertIs(self.c, self.app.queue[0].conf)

    def test_conf_object_consistent_in_new_task(self):
        self.assertEqual(self.app.queue, [])
        t = Task()
        self.assertIsNone(t.conf)
        self.app.add(t)
        self.assertIsNotNone(t.conf)
        self.assertIs(self.c, self.app.queue[0].conf)
        self.assertIs(self.c, t.conf)

    def test_force_options(self):
        self.assertEquals(self.c.runstate.force, self.app.force)
        self.assertFalse(self.c.runstate.force)
        self.assertFalse(self.app.force)

        self.app._force = None
        self.assertFalse(self.app.force)

    def test_default_pool_size(self):
        self.assertIsNotNone(self.c)
        self.assertIsNotNone(self.app.conf)
        self.app._pool_size = None
        self.assertEquals(self.c.runstate.pool_size, self.app.pool_size)

    def tearDown(self):
        self.app.close_pool()


class TestBuildAppMinimalConfig(CommonAppSuite, TestCase):
    @classmethod
    def setUp(self):
        self.app = BuildApp()
        self.app.default_pool = random.choice(['serial', 'thread'])
        self.app.pool_size = 2
        self.c = None

    def tearDown(self):
        self.app.close_pool()


class TestBuildAppAlternateConstructor(CommonAppSuite, TestCase):
    @classmethod
    def setUp(self):
        self.app = BuildApp.new(pool_type=random.choice(['serial', 'thread']),
                                pool_size=None, force=None)
        self.c = None

    def tearDown(self):
        self.app.close_pool()

    # the following tests aren't specifically relevant to this constructor, but
    # rather to the "None" value of the constructor.

    def test_force_setter(self):
        self.assertIsNone(self.app.conf)
        self.assertFalse(self.app.force)
        self.app._force = None
        self.assertFalse(self.app.force)

    def test_pool_size(self):
        self.assertIsNone(self.app.pool_size)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Entry points to view data generated by ``giza.includes`` code. These functions
make it possible to introspect the implicit relationship between source files
and detect unused includes or includes that are only used once, as well as
nested includes.
"""


import os
import logging
import json

import argh

from giza.config.helper import fetch_config

from giza.includes import (included_once, included_recusively,
                           includes_masked, include_files,
                           include_files_unused, changed_includes)

logger = logging.getLogger('giza.operations.includes')

# Helper


def render_for_console(data):
    print(json.dumps(data, indent=3))

# Entry Points


@argh.expects_obj
def recursive(args):
    c = fetch_config(args)

    render_for_console(included_recusively(conf=c))


@argh.expects_obj
def changed(args):
    c = fetch_config(args)

    render_for_console(changed_includes(conf=c))


@argh.expects_obj
def once(args):
    c = fetch_config(args)

    render_for_console(included_once(conf=c))


@argh.expects_obj
def unused(args):
    c = fetch_config(args)

    render_for_console(include_files_unused(conf=c))


@argh.expects_obj
def list(args):
    c = fetch_config(args)

    render_for_console(list(include_files(conf=c).keys()))


@argh.arg('--filter', '-f', default=None, dest="include_mask")
@argh.expects_obj
def graph(args):
    c = fetch_config(args)

    if c.runstate.include_mask is None:
        render_for_console(include_files(conf=c))
    else:
        mask = c.runstate.include_mask
        if mask.startswith(c.paths.source):
            mask = mask[6:]
        elif mask.startswith('/' + c.paths.source):
            mask = mask[7:]

        render_for_console(includes_masked(mask=mask, conf=c))


@argh.expects_obj
def clean(args):
    c = fetch_config(args)

    for fn in include_files_unused(conf=c):
        fn = os.path.join(c.paths.source, fn[1:])
        if os.path.exists(fn):
            os.remove(fn)
            logger.info("removed {0}, which was an unused include file.".format(fn))
        else:
            logger.error('{0} does not exist'.format(fn))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path
import json
import contextlib
import sys
import numbers

import yaml

logger = logging.getLogger('giza.libgiza.config')

if sys.version_info >= (3, 0):
    basestring = str


class ConfigurationError(Exception):
    pass


class OutputError(Exception):
    pass


class ConfigurationBase(object):
    _option_registry = []
    _redacted_keys = ['pass', 'password', 'token', 'key', 'secret']
    _version = 0

    def __init__(self, input_obj=None):
        self._source_fn = None
        self._state = {}
        self.ingest(input_obj)

    def ingest(self, input_obj):
        if input_obj is None:
            return

        input_obj = self._prep_load_data(input_obj)

        # We need deterministic iteration order---"paths" must come before
        # "git", or else we have an uninitialized read from Configuration.paths
        items = list(input_obj.items())
        items.sort(reverse=True)

        for key, value in items:
            try:
                setattr(self, key, value)
            except AttributeError as e:
                m = '{0}({1}) ingestion error with {2} key and {3} value, for {4} obj'
                m = m.format(e, type(e), key, value, type(self))
                logger.error(m)
                raise ConfigurationError(m)

    def _prep_load_data(self, input_obj):
        if isinstance(input_obj, dict):
            return input_obj
        elif not isinstance(input_obj, ConfigurationBase) and os.path.isfile(input_obj):
            self._source_fn = input_obj

            with open(input_obj, 'r') as f:
                if input_obj.endswith('json'):
                    input_obj = json.load(f)
                elif input_obj.endswith('yaml') or input_obj.endswith('yml'):
                    input_obj = yaml.safe_load(f)
                else:
                    logger.error("file {0} has unknown data format".format(input_obj))

            if input_obj is None:
                input_obj = {}
        else:
            msg = 'cannot ingest Configuration obj from {0} object'.format(type(input_obj))
            logger.critical(msg)
            raise TypeError(msg)

        return input_obj

    def __contains__(self, key):
        if key in self.state:
            return True
        elif key.startswith("_") and hasattr(self, key):
            return True
        else:
            return False

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError as e:
            m = 'key "{0}" in configuration object ({1}) is not defined'.format(key, type(self))
            if key in self._option_registry:
                try:
                    return self.state[key]
                except KeyError:
                    raise AttributeError(m, str(e))
            else:
                if not key.startswith('_'):
                    logger.debug(m)
                raise AttributeError(m, str(e))

    def __setattr__(self, key, value):
        if key in self._option_registry:
            self.state[key] = value
        elif key.startswith('_') or key in dir(self):
            object.__setattr__(self, key, value)
        else:
            msg = 'configuration object {0} lacks support for "{1}" value'.format(type(self), key)
            logger.error(msg)
            raise TypeError(msg)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        logger.warning('cannot set state record directly')

    @staticmethod
    def _is_value_type(value):
        acceptable_types = (ConfigurationBase, basestring, list, numbers.Number)

        if isinstance(value, acceptable_types):
            return True
        else:
            return False

    def __repr__(self):
        return str(self.dict())

    def __get_dict_value__(self, v, safe=True):
        if isinstance(v, ConfigurationBase):
            return v.dict(safe)
        elif isinstance(v, list) and len(v) > 0 and isinstance(v[0], ConfigurationBase):
            return [i.dict() for i in v]
        elif isinstance(v, dict):
            sub_d = {}
            for key, value, in v.items():
                sub_d[key] = self.__get_dict_value__(value)
            return sub_d
        elif self._is_value_type(v):
            return v
        else:
            return 'error'

    def dict(self, safe=True):
        d = {}

        for key, value in self.state.items():
            if safe in (True, None):
                if key != "_id" and key.startswith('_'):
                    continue
                elif key in self._redacted_keys:
                    d[key] = 'redacted'
                else:
                    d[key] = self.__get_dict_value__(value, safe)
            elif safe is False:
                d[key] = self.__get_dict_value__(value, safe)

        return d

    def write(self, fn=None, add_version=False):
        if fn is None:
            if self._source_fn is None:
                logger.error('cannot write object to unspecified file.')
                return
            else:
                fn = self._source_fn

        if not isinstance(fn, basestring):
            raise OutputError("unsupported file format: {0}".format(fn))

        if add_version is True and 'v' not in self.state:
            self.state['v'] = self._version

        if fn.endswith('json'):
            with open(fn, 'w') as f:
                json.dump(self.dict(safe=False), f, indent=3, sort_keys=True)
        elif fn.endswith('yaml') or fn.endswith('yml'):
            with open(fn, 'w') as f:
                yaml.safe_dump(self.dict(safe=False), f, default_flow_style=False)
        else:
            raise OutputError("unsupported file format: {0}".format(fn))

    @classmethod
    @contextlib.contextmanager
    def persisting(cls, fn, override=False):
        if not os.path.isfile(fn):
            with open(fn, 'w') as f:
                json.dump({}, f, indent=3)

        if override is False:
            data = cls(fn)
        elif override is True:
            data = cls()
            input_data = data._prep_load_data(fn)
            data.state.update(input_data)

        yield data

        data.write(fn)


class RecursiveConfigurationBase(ConfigurationBase):
    def __init__(self, obj, conf):
        self._conf = None
        self.conf = conf
        super(RecursiveConfigurationBase, self).__init__(obj)

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, value):
        if isinstance(value, ConfigurationBase):
            self._conf = value
        else:
            m = 'invalid configuration object: {0}'.format(value)
            logger.error(m)
            raise TypeError(m)



"""
Provides a simple, lightweight object used to capture data about an error
encountered while checking or validating a resource.
"""
import logging
import multiprocessing
import sys
import threading
import traceback

import giza.libgiza.config

logger = logging.getLogger('giza.libgiza.error')

if sys.version_info >= (3, 0):
    basestring = str

_DEFAULT_ERROR_MESSAGE = "generic error"


class Error(object):
    """
    Collects information about, and represents an error encountered. Use in
    situations where you want to track errors, but don't want to raise
    exceptions (e.g. for continue on error situations.)
    """

    def __init__(self, message=_DEFAULT_ERROR_MESSAGE, include_trace=True, fatal=True):
        self.capture_trace()
        self._include_trace = include_trace
        self._fatal = fatal
        self._payload = None
        self.message = message

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, value):
        if hasattr(self, "_message") and self._message != _DEFAULT_ERROR_MESSAGE:
            raise ValueError("cannot overwrite existing message ({0}) with new message "
                             "({1})".format(self._message, value))
        elif isinstance(value, basestring):
            self._message = value
        else:
            raise TypeError("error message should be a string. cannot set to: "
                            "{0} ({1})".format(value, type(value)))

    @property
    def fatal(self):
        if hasattr(self, "_fatal"):
            return self._fatal
        else:
            return True

    @fatal.setter
    def fatal(self, value):
        if isinstance(value, bool):
            self._fatal = value
        else:
            raise TypeError("fatal option must be a bool. ({0}, {1})".format(value, type(value)))

    @property
    def include_trace(self):
        if hasattr(self, "_include_trace"):
            return self._include_trace
        else:
            return True

    @include_trace.setter
    def include_trace(self, value):
        if isinstance(value, bool):
            self._include_trace = value
        else:
            raise TypeError("include_trace option must be a bool. "
                            "({0}, {1})".format(value, type(value)))

    @property
    def trace(self):
        # we want to drop the last stack frame because it's the assignment to
        # self._trace, which isn't relevant to the error.
        return self._trace[-10:][:-2]

    def capture_trace(self):
        self._trace = traceback.extract_stack()

    @property
    def payload(self):
        if isinstance(self._payload, dict):
            return self._payload
        elif isinstance(self._payload, giza.libgiza.config.ConfigurationBase):
            return self._payload.dict()
        else:
            return {}

    @payload.setter
    def payload(self, value):
        if isinstance(value, (dict, giza.libgiza.config.ConfigurationBase)):
            self._payload = value
        else:
            raise TypeError("error payloads must be dict subclasses or "
                            "giza.libgiza configuration objects, cannot set to: "
                            "{0} ({1})".format(value, type(value)))

    def render_output(self, prefix=""):
        output = []
        if self.fatal:
            output.append("encountered fatal error:")
        else:
            output.append("encountered non-fatal error:")

        output.append("    " + self.message)

        if self.include_trace:
            output.append("traceback:")
            for trace_line in traceback.format_list(self.trace):
                output.extend([ln for ln in trace_line.split("\n") if ln != ''])

        return prefix + ("\n" + prefix).join(output)

    def dict(self):
        return {"message": self.message,
                "payload": self.payload,
                "fatal": self.fatal,
                "trace": [{"file": t[0], "line": t[1], "function": t[2], "operation": t[3]}
                          for t in self.trace]}

    # define correct representations
    def __repr__(self):
        return str(self.dict())

    def __str__(self):
        return self.render_output()

    def __format__(self):
        return self.render_output()


class ErrorCollector(object):
    """
    A class that exists to collect and aggregate ErrorObject instances,
    potentially across threads or process pools.
    """

    def __init__(self, name="error-collector", concurrency_type="thread"):
        self._contains_fatal_error = False
        self.errors = []

        if concurrency_type.startswith("proc"):
            self.lock = multiprocessing.RLock()
        else:
            self.lock = threading.RLock()

        self.name = name

    def __len__(self):
        return len(self.errors)

    @property
    def fatal(self):
        with self.lock:
            return self._contains_fatal_error

    @property
    def count(self):
        with self.lock:
            return len(self.errors)

    @property
    def name(self):
        if hasattr(self, "_name"):
            return self._name
        else:
            return "error-collector"

    @name.setter
    def name(self, value):
        if isinstance(value, basestring):
            self._name = value
        else:
            raise TypeError("name option must be a string. "
                            "({0}, {1})".format(value, type(value)))

    def has_errors(self):
        if self.count > 0:
            return True
        else:
            return False

    def add(self, error):
        if error is None:
            return
        elif isinstance(error, Error):
            with self.lock:
                if error.fatal:
                    self._contains_fatal_error = True

                self.errors.append(error)
        elif isinstance(error, ErrorCollector):
            with self.lock:
                with error.lock:
                    if error.fatal:
                        self._contains_fatal_error = True
                    self.errors.extend(error.errors)
                    error.clear()
        else:
            raise TypeError("can only add ErrorObject and ErrorCollector objects to an "
                            "ErrorCollector. {0} ({1})".format(error, type(error)))

    def clear(self):
        if self.has_errors():
            logger.debug("clearing {0} errors from error group".format(self.count))

        with self.lock:
            self._contains_fatal_error = False
            self.errors = []

    def render_output(self, prefix=""):
        if not self.has_errors():
            return ""
        else:
            with self.lock:
                if self.fatal:
                    output = ["encountered {0} errors. (fatal):".format(self.count)]
                else:
                    output = ["encountered {0} errors:".format(self.count)]

                for num, error in enumerate(self.errors, start=1):
                    output.append("error: {0}".format(num))

                    if prefix == "":
                        output.append(error.render_output("    "))
                    else:
                        output.append(error.render_output(prefix))

                return prefix + ("\n" + prefix).join(output)

    def dict(self):
        if not self.has_errors():
            return {"errors": []}
        else:
            with self.lock:
                return {"errors": [e.dict() for e in self.errors]}

    # define correct representations
    def __repr__(self):
        return str(self.dict())

    def __str__(self):
        return self.render_output()

    def __format__(self):
        return self.render_output()

    def __bool__(self):
        return self.has_errors() and self.fatal

    def __nonzero__(self):
        # Python 2 compat
        return self.__bool__()



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import jinja2

from giza.inheritance import InheritableContentBase
from giza.content.helper import get_all_languages, level_characters

logger = logging.getLogger('giza.content.steps.models')

if sys.version_info >= (3, 0):
    basestring = str


class HeadingMixin(object):
    _default_level = 3

    @property
    def title(self):
        return self.heading

    @title.setter
    def title(self, value):
        self.heading = value

    @property
    def heading(self):
        if self.optional is True:
            return "Optional: " + self.state['heading']
        else:
            return self.state['heading']

    @heading.setter
    def heading(self, value):
        if isinstance(value, dict):
            self.state['heading'] = value['text']
            if ('level' not in self.state and
                    'character' in value and
                    value['character'] in level_characters):

                self.state['level'] = level_characters[value['character']]
            else:
                self.state['level'] = self._default_level
        else:
            self.state['heading'] = value
            if 'level' not in self.state:
                self.state['level'] = self._default_level

    @property
    def level(self):
        if 'level' in self.state:
            return self.state['level']
        else:
            return self._default_level

    @level.setter
    def level(self, value):
        if isinstance(value, basestring):
            if value in level_characters:
                self.state['level'] = level_characters[value]
            else:
                logger.error('{0} is not a valid heading level'.format(value))
        elif isinstance(value, (int, float, complex)):
            self.state['level'] = int(value)
        else:
            logger.error('{0} is not a valid heading level'.format(value))

    @property
    def optional(self):
        if 'optional' in self.state:
            return True
        else:
            return False

    @optional.setter
    def optional(self, value):
        if value is True:
            self.state['optional'] = True
        else:
            self.state['optional'] = False


class ActionMixin(object):
    @property
    def code(self):
        return self.state['code']

    @code.setter
    def code(self, value):
        if isinstance(value, list):
            self.state['code'] = value
        else:
            self.state['code'] = value.split('\n')

    @property
    def language(self):
        return self.state['language']

    @language.setter
    def language(self, value):
        if value in get_all_languages():
            self.state['language'] = value
        else:
            m = '{0} is not a supported language'.format(value)
            logger.error(m)
            TypeError(m)

    @property
    def copyable(self):
        return self.state.get('copyable', True)

    @copyable.setter
    def copyable(self, value):
        if not isinstance(value, bool):
            raise TypeError('"copyable" expects boolean')

        self.state['copyable'] = value


class StepData(HeadingMixin, InheritableContentBase):
    _default_level = 3

    @property
    def number(self):
        if not hasattr(self, '_number'):
            self._number = None

        return self._number

    @number.setter
    def number(self, value):
        if not hasattr(self, '_number'):
            self._number = None

        if isinstance(value, (int, float, complex)):
            self._number = int(value)
            self.state['number'] = self._number
        else:
            raise TypeError

    @property
    def stepnum(self):
        return self.number

    @stepnum.setter
    def stepnum(self, value):
        self.number = value

    @property
    def action(self):
        return self.state['action']

    @action.setter
    def action(self, value):
        if isinstance(value, ActionContent):
            self.state['action'] = [value]
        elif isinstance(value, dict):
            self.state['action'] = [ActionContent(value, self.conf)]
        elif isinstance(value, list):
            actions = []
            for item in value:
                if isinstance(item, ActionContent):
                    actions.append(item)
                else:
                    actions.append(ActionContent(item, self.conf))
            self.state['action'] = actions


class ActionContent(HeadingMixin, ActionMixin, InheritableContentBase):
    _option_registry = ['copyable', 'pre', 'post', 'content']

    def render(self):
        if self.replacement and 'code' in self:
            super(ActionContent, self).render()
            code_block = '\n'.join(self.code)

            for attempt in range(10):
                if "{{" in code_block:
                    template = jinja2.Template(code_block)
                    code_block = template.render(**self.replacement)
                    if "{{" not in code_block:
                        self.code = code_block
                        return
                elif attempt == 0:
                    return
                else:
                    self.code = code_block
                    return



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Create and manipulate "packages" of build artifacts that can be used to quickly
share and deploy artifacts and build environment data.
"""

import logging
import datetime
import os
import tarfile
import contextlib
from six.moves import urllib

import argh
import giza.libgiza.app

from giza.config.helper import fetch_config
from giza.tools.files import safe_create_directory, FileNotFoundError
from giza.operations.deploy import deploy_tasks

logger = logging.getLogger('giza.operations.packaging')

# Helper


def package_filename(target, conf):
    archive_path = os.path.join(conf.paths.projectroot, conf.paths.buildarchive)

    fn = [conf.project.name]

    if target is not None:
        tag = ''.join([i if i not in ['push', 'stage'] else '' for i in target.split('-')])
        if tag != '':
            fn.append(tag)

    fn.extend([conf.git.branches.current,
               datetime.datetime.utcnow().strftime('%s'),
               conf.git.commit[:8]])

    fn = os.path.join(archive_path, '-'.join(fn) + '.tar.gz')

    return fn


def create_archive(files_to_archive, tarball_name):
    # ready to write the tarball

    safe_create_directory(os.path.dirname(tarball_name))

    with tarfile.open(tarball_name, 'w:gz') as t:
        for fn, arc_fn in files_to_archive:
            t.add(name=fn, arcname=arc_fn)

# Worker Functions


def create_package(target, conf):
    logger.info('creating package for target "{0}"'.format(target))

    if target is None:
        pconf = conf.system.files.data.push[0]
        target = pconf['target']
    else:
        pconf = dict((item['target'], item) for item in conf.system.files.data.push)[target]

    files_to_archive = []

    if conf.project.branched is True:
        artifacts = (os.path.join(conf.paths.output,
                                  conf.git.branches.current),
                     conf.git.branches.current)
    else:
        artifacts = (os.path.join(conf.paths.projectroot,
                                  conf.paths.output,
                                  pconf['paths']['local']),
                     os.path.split(pconf['paths']['local'])[-1])

    files_to_archive.append(artifacts)

    if 'static' in pconf['paths']:
        files_to_archive.extend([
            (os.path.join(conf.paths.projectroot, conf.paths.public, path), path)
            for path in pconf['paths']['static']
            if os.path.exists(os.path.join(conf.paths.projectroot, conf.paths.public, path))
        ])

    archive_fn = package_filename(target, conf)

    create_archive(files_to_archive, archive_fn)

    logger.info('wrote build package to: {0}'.format(archive_fn))


def extract_package(conf):
    path = conf.runstate.package_path

    if conf.runstate.package_path.startswith('http'):
        path = fetch_package(path, conf)
    elif os.path.isfile(conf.runstate.package_path):
        path = conf.runstate.package_path
    else:
        m = "package {0} does not exist".format(conf.runstate.package_path)
        logger.critical(m)
        raise FileNotFoundError(m)

    with tarfile.open(path, "r:gz") as t:
        t.extractall(os.path.join(conf.paths.projectroot, conf.paths.public))


def fetch_package(path, conf):
    if path.startswith('http'):

        local_path = path.split('/')[-1]

        tar_path = os.path.join(conf.paths.projectroot,
                                conf.paths.buildarchive,
                                local_path)

        if not os.path.exists(tar_path):
            with contextlib.closing(urllib.request.urlopen(path)) as u:
                with open(tar_path, 'w') as f:
                    f.write(u.read())
            logger.info('downloaded {0}'.format(local_path))
        else:
            logger.info('{0} exists locally, not downloading.'.format(local_path))

        return tar_path
    elif os.path.isfile(path):
        return path
    else:
        msg = 'no archive named "{0}" exists'.format(path)
        logger.error(msg)
        raise FileNotFoundError(msg)

# Command Entry Points


@argh.arg('--target', '-t', nargs="*", dest='push_targets')
@argh.expects_obj
def create(args):
    conf = fetch_config(args)

    for target in conf.runstate.push_targets:
        create_package(target, conf)


@argh.arg('--path', dest='package_path')
@argh.expects_obj
def unwind(args):
    conf = fetch_config(args)

    conf.runstate.package_path = fetch_package(conf.runstate.package_path, conf)
    logger.info('extracting package: ' + conf.runstate.package_path)
    extract_package(conf)
    logger.info('extracted package')


@argh.arg('--path', dest='package_path')
@argh.arg('--target', '-t', nargs=1, dest='push_targets')
@argh.arg('--dry-run', '-d', action='store_true', dest='dry_run')
@argh.expects_obj
def deploy(args):
    conf = fetch_config(args)

    conf.runstate.package_path = fetch_package(conf.runstate.package_path, conf)

    logger.info('extracting package: ' + conf.runstate.package_path)
    extract_package(conf)
    logger.info('extracted package')

    app = giza.libgiza.app.BuildApp.new(pool_type=conf.runstate.runner,
                                        pool_size=conf.runstate.pool_size,
                                        force=conf.runstate.force)

    logger.info('beginning deploy now.')
    deploy_tasks(conf, app)

    if conf.runstate.dry_run is False:
        app.run()


@argh.arg('--path', dest='package_path')
@argh.expects_obj
def fetch(args):
    conf = fetch_config(args)

    if conf.runstate.package_path.startswith('http'):
        fetch_package(conf.runstate.pacage_path, conf)
    else:
        logger.error('{0} is not a url'.format(conf.runstate.package_path))
        raise SystemExit



"""
Unittests of error tracker and collectors. These objects provide a way to track
errors during validation processes, without raising expections until the end of
an aggregated process, without throwing away information about the error itself.
"""
import sys
import unittest
import multiprocessing
import threading

import giza.libgiza.error

if sys.version_info >= (3, 0):
    basestring = str


class TestErrorConstruction(unittest.TestCase):
    def test_create_without_arguments_have_expected_values(self):
        error = giza.libgiza.error.Error()

        self.assertIsNone(error._payload)
        self.assertEquals({}, error.payload)

        self.assertTrue(error.fatal)
        self.assertTrue(error.include_trace)
        self.assertEquals("generic error", error.message)

    def test_create_with_positional_arguments_have_values_that_persist(self):
        msg = "a new message"
        error = giza.libgiza.error.Error(msg, False, False)

        self.assertIsNone(error._payload)
        self.assertEquals({}, error.payload)

        self.assertFalse(error.fatal)
        self.assertFalse(error.include_trace)
        self.assertEquals(msg, error.message)

    def test_create_with_keyword_arguments_have_values_that_persist(self):
        msg = "a new message"
        error = giza.libgiza.error.Error(message=msg, include_trace=False, fatal=False)

        self.assertIsNone(error._payload)
        self.assertEquals({}, error.payload)

        self.assertFalse(error.fatal)
        self.assertFalse(error.include_trace)
        self.assertEquals(msg, error.message)


class TestErrorObject(unittest.TestCase):
    def setUp(self):
        self.error = giza.libgiza.error.Error()

    def test_message_setter_only_accepts_strings_and_raises_type_error_otherwise(self):
        self.assertEquals("generic error", self.error.message)

        for value in [1.02, True, False, None, 42, {"a": 1}, (1, 2), [True, False], {}, []]:
            with self.assertRaises(TypeError):
                self.error.message = value
            self.assertEquals("generic error", self.error.message)

    def test_message_setter_raises_value_error_after_setting_once(self):
        self.assertEquals("generic error", self.error.message)

        self.error.message = "foo"
        self.assertEquals("foo", self.error.message)

        with self.assertRaises(ValueError):
            self.error.message = "bar"

    def test_fatal_setter_returns_true_value_by_default(self):
        self.assertTrue(self.error.fatal)
        del self.error._fatal
        self.assertFalse(hasattr(self.error, "_fatal"))
        self.assertTrue(self.error.fatal)

    def test_fatal_setter_persists_bool_values(self):
        self.assertTrue(self.error.fatal)
        del self.error._fatal
        self.assertFalse(hasattr(self.error, "_fatal"))

        self.error.fatal = False
        self.assertFalse(self.error.fatal)
        self.error.fatal = True
        self.assertTrue(self.error.fatal)
        self.assertTrue(hasattr(self.error, "_fatal"))

    def test_fatal_setter_raises_type_error_when_setting_to_non_boolean(self):
        for value in [1.02, "true", "false", "t", "f", "other string",
                      42, {"a": 1}, (1, 2), [True, False], {}, []]:
            with self.assertRaises(TypeError):
                self.error.fatal = value

    def test_include_trace_setter_returns_true_value_by_default(self):
        self.assertTrue(self.error.include_trace)
        del self.error._include_trace
        self.assertFalse(hasattr(self.error, "_include_trace"))
        self.assertTrue(self.error.include_trace)

    def test_include_trace_setter_raises_type_error_when_setting_to_non_boolean(self):
        for value in [1.02, "true", "false", "t", "f", "other string",
                      42, {"a": 1}, (1, 2), [True, False], {}, []]:
            with self.assertRaises(TypeError):
                self.error.include_trace = value

    def test_include_trace_setter_persists_bool_values(self):
        self.assertTrue(self.error.include_trace)
        del self.error._include_trace
        self.assertFalse(hasattr(self.error, "_include_trace"))

        self.error.include_trace = False
        self.assertFalse(self.error.include_trace)
        self.error.include_trace = True
        self.assertTrue(self.error.include_trace)
        self.assertTrue(hasattr(self.error, "_include_trace"))

    def test_payload_attribute_always_returns_dict(self):
        self.assertIsInstance(self.error.payload, dict)
        self.assertIsNone(self.error._payload)
        self.assertEquals(0, len(self.error.payload))

        conf = giza.libgiza.config.ConfigurationBase()
        conf._option_registry = ["a"]
        conf.a = 1

        for obj in [{"a": 1}, {1: True}, {}, conf,
                    giza.libgiza.config.ConfigurationBase()]:
            self.error.payload = obj
            self.assertIs(obj, self.error._payload)
            self.assertIsInstance(self.error.payload, dict)
            if isinstance(obj, giza.libgiza.config.ConfigurationBase):
                self.assertEquals(len(obj.state), len(self.error.payload))
            else:
                self.assertEquals(len(obj), len(self.error.payload))

    def test_paylod_setter_raises_type_error_when_setting_incorret_type(self):
        for value in [1.02, "true", "false", "t", "f", "other string",
                      42, (1, 2), [True, False], []]:
            with self.assertRaises(TypeError):
                self.error.payload = value

    def test_render_output_returns_string(self):
        output = self.error.render_output()
        self.assertTrue(len(output) > 1)
        self.assertIsInstance(output, basestring)

    def test_render_output_always_prefixes_value(self):
        for prefix in [" ", "   ", "   ", "--", "---", "--->"]:
            output = self.error.render_output(prefix=prefix)
            for ln in output.split("\n"):
                if ln == "":
                    continue

                self.assertTrue(ln.startswith(prefix))

    def test_dict_struct_contains_correct_keys_and_types(self):
        obj = self.error.dict()
        for key in ("message", "payload", "fatal", "trace"):
            self.assertTrue(key in obj)

        self.assertIsInstance(obj["message"], basestring)
        self.assertIsInstance(obj["payload"], dict)
        self.assertIsInstance(obj["fatal"], bool)
        self.assertIsInstance(obj["trace"], list)
        for trace in obj['trace']:
            self.assertIsInstance(trace, dict)
            for key in ("file", "line", "function", "operation"):
                self.assertTrue(key in trace)
                if key == "line":
                    self.assertIsInstance(trace[key], int)
                elif trace[key] is None:
                    # for pypy, the output of traceback is different.
                    continue
                else:
                    self.assertIsInstance(trace[key], basestring)


class CollectorChecks(object):
    def test_emtpy_object_reports_does_not_has_errors(self):
        self.assertFalse(self.collector.fatal)
        self.assertFalse(self.collector.has_errors())

    def test_add_error_modifies_state_of_error_collector(self):
        err = giza.libgiza.error.Error()
        self.assertFalse(self.collector.has_errors())
        self.collector.add(err)
        self.assertTrue(self.collector.has_errors())
        self.assertTrue(self.collector.fatal)
        self.assertTrue(1, len(self.collector))

    def test_add_error_method_raises_type_error_for_invalid_error_types(self):
        for value in [1.02, True, False, 42, {"a": 1}, (1, 2),
                      object(), Exception(), self, TypeError, [True, False], {}, []]:
            with self.assertRaises(TypeError):
                self.collector.add(value)

    def test_add_collector_to_collector_causes_second_collector_to_be_absorbed(self):
        self.assertFalse(self.collector.has_errors())
        sub_collector = giza.libgiza.error.ErrorCollector()
        for _ in range(3):
            self.collector.add(giza.libgiza.error.Error())
            sub_collector.add(giza.libgiza.error.Error())

        self.assertTrue(self.collector.has_errors())
        self.assertEquals(3, len(self.collector))
        self.assertTrue(self.collector.has_errors())
        self.assertEquals(3, len(sub_collector))

        self.collector.add(sub_collector)
        self.assertTrue(self.collector.has_errors())
        self.assertEquals(6, len(self.collector))
        self.assertFalse(sub_collector.has_errors())

    def test_collector_with_non_fatal_errors_render_output_returns_string(self):
        for _ in range(3):
            self.collector.add(giza.libgiza.error.Error(fatal=False))

        output = self.collector.render_output()
        self.assertTrue(len(output) > 1)
        self.assertIsInstance(output, basestring)

    def test_collector_with_fatal_errors_render_output_returns_string(self):
        for _ in range(3):
            self.collector.add(giza.libgiza.error.Error(fatal=True))

        output = self.collector.render_output()
        self.assertTrue(len(output) > 1)
        self.assertIsInstance(output, basestring)
        self.assertTrue("(fatal)" in output)

    def test_render_output_always_prefixes_value(self):
        for _ in range(3):
            self.collector.add(giza.libgiza.error.Error())

        for prefix in [" ", "   ", "   ", "--", "---", "--->"]:
            output = self.collector.render_output(prefix=prefix)
            for ln in output.split("\n"):
                if ln == "":
                    continue

                self.assertTrue(ln.startswith(prefix))

    def test_collector_render_string_returns_emtpy_string_without_errors(self):
        self.assertFalse(self.collector.has_errors())
        self.assertEquals("", self.collector.render_output())

    def test_dict_output_from_empty_collector_is_well_typed(self):
        empty_output = self.collector.dict()
        self.assertTrue("errors" in empty_output)
        self.assertIsInstance(empty_output["errors"], list)
        self.assertEquals(0, len(empty_output["errors"]))

    def test_dict_output_from_collector_is_well_typed(self):
        for _ in range(3):
            self.collector.add(giza.libgiza.error.Error(fatal=True))

        output = self.collector.dict()
        self.assertTrue("errors" in output)
        self.assertIsInstance(output["errors"], list)
        self.assertEquals(3, len(output["errors"]))
        for error in output["errors"]:
            self.assertIsInstance(error, dict)

            for key in ("message", "payload", "fatal", "trace"):
                self.assertTrue(key in error)

            self.assertIsInstance(error["message"], basestring)
            self.assertIsInstance(error["payload"], dict)
            self.assertIsInstance(error["fatal"], bool)
            self.assertIsInstance(error["trace"], list)

            for trace in error['trace']:
                self.assertIsInstance(trace, dict)
                for key in ("file", "line", "function", "operation"):
                    self.assertTrue(key in trace)
                    if key == "line":
                        self.assertIsInstance(trace[key], int)
                    elif trace[key] is None:
                        # for pypy, the output of traceback is different.
                        continue
                    else:
                        self.assertIsInstance(trace[key], basestring)


class TestErrorCollectorProcessLock(CollectorChecks, unittest.TestCase):
    def setUp(self):
        self.collector = giza.libgiza.error.ErrorCollector(name="process-collector",
                                                           concurrency_type="process")

    def test_collector_reports_expected_name(self):
        self.assertEquals("process-collector", self.collector.name)

    def test_collector_lock_uses_correct_type(self):
        self.assertEquals(type(self.collector.lock), type(multiprocessing.RLock()))


class TestErrorCollectorThreadLock(CollectorChecks, unittest.TestCase):
    def setUp(self):
        self.collector = giza.libgiza.error.ErrorCollector(name="thread-collector",
                                                           concurrency_type="thread")

    def test_collector_reports_expected_name(self):
        self.assertEquals("thread-collector", self.collector.name)

    def test_collector_lock_uses_correct_type(self):
        self.assertEquals(type(self.collector.lock), type(threading.RLock()))


class TestErrorCollector(unittest.TestCase):
    def setUp(self):
        self.collector = giza.libgiza.error.ErrorCollector()

    def test_default_name_behavior(self):
        self.assertTrue(hasattr(self.collector, "_name"))
        self.assertEquals("error-collector", self.collector.name)

        del self.collector._name
        self.assertFalse(hasattr(self.collector, "_name"))
        self.assertEquals("error-collector", self.collector.name)

    def test_name_setter_raises_type_error_for_non_string_values(self):
        for value in [1.02, True, False, None, 42, {"a": 1}, (1, 2), [True, False], {}, []]:
            with self.assertRaises(TypeError):
                self.collector.name = value



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from rstcloth.rstcloth import RstCloth

logger = logging.getLogger('giza.content.glossary.views')


def render_glossary(terms):
    r = RstCloth()

    r.directive(name="glossary", fields=[("sorted", "")])
    r.newline()

    for term in terms.ordered_content():
        r.definition(term.term, term.definition, wrap=False, indent=3)

    return r



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import contextlib

logger = logging.getLogger('giza.libgit')

try:
    import pygit2
except ImportError:
    pygit2 = None
    import giza.libgiza.git


class GitError(Exception):
    pass


class GitRepo(object):
    def __init__(self, path=None):
        """
        :param string path: Optional. Defines a the path of the git
           repository. If not specified, defaults to the current working
           directory.
        """
        if pygit2 is None:
            logger.warning("without pylibgit2 installed, falling back "
                           "to wrapped git repo implementation.")
            return giza.libgiza.git.GitRepo(path)

        if path is None:
            self.path = os.getcwd()
        else:
            self.path = path

        self.path = pygit2.discover_repository(self.path)
        self.repo = pygit2.Repository(self.path)

        logger.debug("created git repository management object for {0}".format(self.path))

    def cmd(self, *args):
        raise NotImplementedError

    def remotes(self):
        return [r.name for r in self.repo.remotes]

    def author_email(self, sha=None):
        commit = self.repo.get(sha, None)

        if commit is None or not isinstance(commit, pygit2.Commit):
            return ""
        else:
            return commit.author

    def branch_exists(self, name):
        return name in self.repo.listall_branches()

    def branch_file(self, path, branch="master"):
        ref = self.repo.lookup_branch(branch)
        head = self.repo.lookup_reference(ref.name)
        commit = self.repo.get(head.target)

        return self.repo[commit.tree[path].id]

    def checkout(self, ref):
        self.repo.checkout(refnmae=ref)

        return True

    def checkout_branch(self, name, tracking=None):
        pass

    def remote_branch(self, name, force=False):
        pass

    def rebase(self, onto):
        pass

    def merge(self, branch):
        pass

    def hard_reset(self, ref='head'):
        pass

    def fetch(self, remote='origin'):
        pass

    def update(self):
        pass

    def pull(self, remote='origin', branch='master'):
        pass

    def current_branch(self):
        pass

    def sha(self, ref='HEAD'):
        pass

    def clone(self, remote, repo_path=None, branch=None):
        pass

    def am(self, patches, repo=None, sign=False):
        pass

    @contextlib.contextmanager
    def branch(self, name):
        pass



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import os
import tarfile
import tempfile
import contextlib

import argh
import giza.libgiza.task

from giza.libgiza.app import BuildApp
from sphinx.application import Sphinx, ENV_PICKLE_FILENAME
from sphinx.builders.html import get_stable_hash

from giza.config.helper import fetch_config, get_builder_jobs
from giza.config.sphinx_config import available_sphinx_builders
from giza.operations.packaging import fetch_package
from giza.tools.files import safe_create_directory, FileNotFoundError

logger = logging.getLogger('giza.operations.build_env')

# Helpers


@contextlib.contextmanager
def cd(path):
    cur_dir = os.getcwd()

    os.chdir(path)

    yield

    os.chdir(cur_dir)


def is_git_dir(path):
    git_dir = ''.join([os.path.sep, '.git', os.path.sep])
    if git_dir in path:
        return True
    else:
        return False


def extract_package_at_root(path, conf):
    with cd(conf.paths.projectroot):
        with tarfile.open(path, "r:gz") as t:
            t.extractall()


def get_existing_builders(conf):
    return [b
            for b in available_sphinx_builders()
            if os.path.isdir(os.path.join(conf.paths.projectroot, conf.paths.branch_output, b))]


def env_package_worker(args, conf):
    # used by the make interface
    package_build_env(args.builder, args.editions_to_build, args.languages_to_build, conf)

# Core Workers


def package_build_env(builders, editions, languages, conf):
    arc_fn = '-'.join(['cache',
                       conf.project.name,
                       conf.git.branches.current,
                       datetime.datetime.utcnow().strftime('%s'),
                       conf.git.commit[:8]]) + ".tar.gz"
    archive_path = os.path.join(conf.paths.buildarchive, arc_fn)
    safe_create_directory(conf.paths.buildarchive)

    existing_archives = os.listdir(conf.paths.buildarchive)

    for arc in existing_archives:
        if conf.git.commit[:8] in arc:
            m = 'archive "{0}" exists for current git hash, not recreating'
            logger.warning(m.format(archive_path))
            return

    logger.debug("no archive for commit '{0}' continuing".format(conf.git.commit))

    with cd(conf.paths.projectroot):
        files_to_archive = set()

        for ((edition, language, builder), (rconf, sconf)) in get_builder_jobs(conf):
            files_to_archive.add(rconf.paths.branch_source)
            files_to_archive.add(os.path.join(rconf.paths.branch_output,
                                              sconf.build_output))
            files_to_archive.add(os.path.join(rconf.paths.branch_output,
                                              '-'.join(('doctrees', sconf.build_output))))
            files_to_archive.add(rconf.system.dependency_cache_fn)

        files_to_archive = list(files_to_archive)
        logger.info('prepped build cache archive. writing file now.')

        for fn in files_to_archive:
            if not os.path.exists(fn):
                raise FileNotFoundError(fn)

        try:
            with tarfile.open(archive_path, 'w:gz') as t:
                for fn in files_to_archive:
                    t.add(fn, exclude=is_git_dir)
            logger.info("created build-cache archive: " + archive_path)
        except Exception as e:
            os.remove(archive_path)
            logger.critical("failed to create archive: " + archive_path)
            logger.error(e)


def fix_build_env(builder, conf):
    """
    Given a builder name and the conf object, this function fixes the build
    artifacts for the current build to prevent a full rebuild. Currently
    re-pickles the environment and dumps the ``.buildinfo`` file in the build
    directory with the correct hashes.
    """

    fn = os.path.join(conf.paths.projectroot, conf.paths.branch_output, builder, '.buildinfo')
    logger.info('updating cache for: ' + builder)

    if not os.path.isfile(fn):
        return

    doctree_dir = os.path.join(conf.paths.projectroot,
                               conf.paths.branch_output,
                               "doctrees-" + builder)

    sphinx_app = Sphinx(srcdir=os.path.join(conf.paths.projectroot,
                                            conf.paths.branch_output, "source"),
                        confdir=conf.paths.projectroot,
                        outdir=os.path.join(conf.paths.projectroot,
                                            conf.paths.branch_output, builder),
                        doctreedir=doctree_dir,
                        buildername=builder,
                        status=tempfile.NamedTemporaryFile(),
                        warning=tempfile.NamedTemporaryFile())

    sphinx_app.env.topickle(os.path.join(doctree_dir, ENV_PICKLE_FILENAME))

    with open(fn, 'r') as f:
        lns = f.readlines()
        tags_hash_ln = None
        for ln in lns:
            if ln.startswith('tags'):
                tags_hash_ln = ln
                break

        if tags_hash_ln is None:
            tags_hash_ln = 'tags: ' + get_stable_hash(sorted(sphinx_app.tags))

    with open(fn, 'w') as f:
        config_dict = dict((name, sphinx_app.config[name])
                           for (name, desc) in sphinx_app.config.values.items()
                           if desc[1] == 'html')

        f.write('# Sphinx build info version 1')
        f.write('\n\n')  # current format requires an extra line here.
        f.write('config: ' + get_stable_hash(config_dict))
        f.write('\n')
        f.write(tags_hash_ln)
        f.write('\n')

# Task Creators


def fix_build_env_tasks(builders, conf):
    tasks = []

    message = "fix up sphinx environment for builder '{0}'"
    for builder in builders:
        t = giza.libgiza.task.Task(job=fix_build_env,
                                   args=(builder, conf),
                                   target=True,
                                   dependency=None,
                                   description=message.format(builder))
        tasks.append(t)

    return tasks

# Entry Points


@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*', dest='languages_to_build')
@argh.arg('--builder', '-b', nargs='*', default='html')
@argh.expects_obj
def package(args):
    conf = fetch_config(args)

    package_build_env(builders=conf.runstate.builder,
                      editions=conf.runstate.editions_to_build,
                      languages=conf.runstate.languages_to_build,
                      conf=conf)


@argh.arg('--path', '-p', default=None, dest='_path')
@argh.expects_obj
def extract(args):
    conf = fetch_config(args)

    with BuildApp.new(pool_type=conf.runstate.runner,
                      pool_size=conf.runstate.pool_size,
                      force=conf.runstate.force).context() as app:
        path = fetch_package(conf.runstate._path, conf)
        extract_package_at_root(path, conf)

        builders = get_existing_builders(conf)
        app.extend_queue(fix_build_env_tasks(builders, conf))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Alternate implementation of the system that fetches intersphinx inventories from
other Sphinx sites. Implemented separately to avoid using Sphinx so that that
Giza can check and update inventories once per day rather than once per
build run. Useful for minimizing the start-up time for ``sphinx-build`` and
useful for reducing redundant work in parallel build situations.
"""

import email.utils
import logging
import os
import time
from six.moves import urllib

import giza.libgiza.task

from giza.tools.files import verbose_remove, safe_create_directory

logger = logging.getLogger('giza.content.intersphinx')

MAX_AGE = 60 * 60 * 24 * 1  # One day
TIMEOUT_SECONDS = 5


def download(path, url, conf):
    try:
        mtime = os.stat(path).st_mtime
    except OSError:
        mtime = -1

    now = time.time()
    if now < (mtime + MAX_AGE):
        logger.debug('Intersphinx file still young: %s', url)
        return

    request = urllib.request.Request(url, headers={
        'If-Modified-Since': email.utils.formatdate(mtime)
    })

    safe_create_directory(os.path.dirname(path))

    try:
        response = urllib.request.urlopen(request, timeout=TIMEOUT_SECONDS)
        with open(path, 'wb') as f:
            f.write(response.read())
    except urllib.error.HTTPError as err:
        if err.code == 304:
            logger.debug('Not modified: %s', url)
            return
        logger.error('Error downloading %s: Got %d', url, err.code)
    except urllib.error.URLError as err:
        logger.error('Error downloading %s: %s', url, str(err))


def intersphinx_tasks(conf):
    if 'intersphinx' not in conf.system.files.data:
        return

    tasks = []
    for i in conf.system.files.data.intersphinx:
        try:
            f = os.path.join(conf.paths.projectroot,
                             conf.paths.output, i.path)

            s = i.url + 'objects.inv'
        except AttributeError:
            f = os.path.join(conf.paths.projectroot,
                             conf.paths.output, i['path'])

            s = i['url'] + 'objects.inv'

        description = 'download intersphinx inventory from {0}'.format(s)
        tasks.append(giza.libgiza.task.Task(job=download,
                                            args=(f, s, conf),
                                            target=f,
                                            dependency=None,
                                            description=description))
        logger.debug('added job for {0}'.format(s))

    return tasks


def intersphinx_clean(conf):
    tasks = []

    for inv in conf.system.files.data.intersphinx:
        try:
            fn = os.path.join(conf.paths.projectroot,
                              conf.paths.output, inv.path)
        except AttributeError:
            fn = os.path.join(conf.paths.projectroot,
                              conf.paths.output, inv['path'])

        if os.path.exists(fn):
            t = giza.libgiza.task.Task(job=verbose_remove,
                                       args=[fn])
            tasks.append(t)

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Diagnostic operation that displays the majority of the current configuration
object.
"""

import json
import logging
import sys

import giza
import argh
import giza.libgiza

from giza.config.helper import fetch_config

logger = logging.getLogger('giza.operations.configuration')


@argh.named('version')
@argh.expects_obj
def report_version(args):
    "Returns current version of giza"

    print("giza: " + giza.__version__)
    print("giza.libgiza: " + giza.libgiza.__version__)


@argh.arg('config_target', nargs='*')
@argh.arg('--simple', action='store_true')
@argh.arg('--conf_path', '-c')
@argh.arg('--edition', '-e')
@argh.arg('--language', '-l')
@argh.named('config')
@argh.expects_obj
def render_config(args):
    """Returns configuration object for diagnostic purposes."""

    c = fetch_config(args)

    # the following values are rendered lazily. we list them here so that the
    # final object will be useful to inspect.
    [c.git.commit, c.paths.public, c.git.branches.current,
     c.git.branches.manual, c.git.branches.published,
     c.paths.branch_output, c.paths.buildarchive,
     c.paths.branch_source, c.paths.branch_staging,
     c.paths.branch_images, c.paths.branch_includes,
     c.version.published, c.version.stable, c.version.upcoming,
     c.project.edition, c.deploy, c.paths.global_config,
     c.project.branched, c.system.dependency_cache,
     c.system.dependency_cache_fn, c.paths.public_site_output,
     c.system.content, c.runstate.runner, c.runstate.force,
     c.system.files, c.system.files.paths, c.system.files.data,
     c.system.files.data.integration, c.paths.htaccess]

    # Print out everything
    if not args.config_target:
        print(json.dumps(c.dict(), indent=2))
        return

    query = args.config_target[0]
    cursor = c

    for segment in query.split('.'):
        try:
            if hasattr(cursor, segment):
                cursor = getattr(cursor, segment)
            else:
                cursor = cursor[segment]
        except (KeyError, TypeError):
            logger.fatal('No key "%s" in configuration', query)
            sys.exit(1)

    # Convert a giza.config.system.SystemConfigData object into a dict
    if hasattr(cursor, 'dict'):
        cursor = cursor.dict()

    if args.simple:
        try:
            result = ' '.join(cursor)
        except TypeError:
            logger.fatal('Cannot represent %s with --simple', query)
            sys.exit(1)
    else:
        result = json.dumps(cursor, indent=2)

    print(result)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.libgiza.config import ConfigurationBase

logger = logging.getLogger('giza.config.credentials')


def get_credentials_skeleton():
    return {
        'jira': {
            'username': None,
            'password': None,
            'url': None,
        },
        'corp': {
            'username': None,
            'password': None,
            'seed': None,
        },
        'github': {
            'username': None,
            'password': None,
            'token': None,
        },
        'aws': {
            'key': None,
            'secret': None
        },
        's3': {
            'key': None,
            'secret': None
        }
    }


class CredentialsConfig(ConfigurationBase):
    _option_registry = []

    @property
    def jira(self):
        return self.state['jira']

    @jira.setter
    def jira(self, value):
        self.state['jira'] = JiraCredentialsConfig(value)

    @property
    def corp(self):
        return self.state['corp']

    @corp.setter
    def corp(self, value):
        self.state['corp'] = CorpCredentialsConfig(value)

    @property
    def github(self):
        return self.state['github']

    @github.setter
    def github(self, value):
        self.state['github'] = GithubCredentialsConfig(value)

    @property
    def aws(self):
        return self.state['aws']

    @aws.setter
    def aws(self, value):
        self.state['aws'] = AwsCredentialsConfig(value)

    @property
    def s3(self):
        return self.state['s3']

    @s3.setter
    def s3(self, value):
        self.state['s3'] = AwsCredentialsConfig(value)

    @property
    def rhn(self):
        return self.state['rhn']

    @rhn.setter
    def rhn(self, value):
        self.state['rhn'] = RhnCredentialsConfig(value)


class JiraCredentialsConfig(ConfigurationBase):
    _option_registry = ['username', 'password']

    @property
    def url(self):
        if 'url' in self.state:
            return self.state['url']
        else:
            logger.error("jira url is not specified.")
            return ""

    @url.setter
    def url(self, value):
        value = value.strip()

        if value.startswith('http'):
            self.state['url'] = value
            if value[4] != 's':
                logger.warning("jira url is not https.")
        else:
            logger.error("invalid jira url")
            raise TypeError


class CorpCredentialsConfig(ConfigurationBase):
    _option_registry = ['username', 'password', 'seed']


class GithubCredentialsConfig(ConfigurationBase):
    _option_registry = ['username', 'password', 'token']


class AwsCredentialsConfig(ConfigurationBase):
    _option_registry = ['key', 'secret']


class RhnCredentialsConfig(ConfigurationBase):
    _option_registry = ['username', 'password']



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from giza.libgiza.config import ConfigurationBase


class AssetsConfig(ConfigurationBase):
    _option_registry = ['path', 'branch', 'repository', 'commit']

    @property
    def generate(self):
        return self.state['generate']

    @generate.setter
    def generate(self, value):
        if isinstance(value, list):
            self.state['generate'] = value
        else:
            raise TypeError



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main controlling operations for running Sphinx builds.
"""

import itertools
import logging
import argh

from giza.config.helper import fetch_config, get_builder_jobs, get_restricted_builder_jobs
from giza.libgiza.app import BuildApp
from giza.libgiza.task import Task

from giza.content.robots import robots_txt_tasks
from giza.content.images.tasks import image_tasks
from giza.content.intersphinx import intersphinx_tasks
from giza.content.hash import hash_tasks
from giza.content.source import source_tasks, latex_image_transfer_tasks
from giza.content.dependencies import refresh_dependency_tasks, dump_file_hash_tasks
from giza.content.sphinx import sphinx_tasks, output_sphinx_stream
from giza.content.post.sphinx import finalize_sphinx_build
from giza.content.migrations import migration_tasks
from giza.content.assets import assets_tasks

from giza.tools.timing import Timer
from functools import reduce

logger = logging.getLogger('giza.operations.sphinx')


@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*', dest='languages_to_build')
@argh.arg('--builder', '-b', nargs='*', default='html')
@argh.arg('--serial_sphinx', action='store_true')
@argh.named('sphinx')
@argh.expects_obj
def main(args):
    """
    Use Sphinx to generate build artifacts. Can generate artifacts for multiple
    output types, content editions and translations.
    """
    conf = fetch_config(args)

    app = BuildApp.new(pool_type=conf.runstate.runner,
                       pool_size=conf.runstate.pool_size,
                       force=conf.runstate.force)

    with Timer("full sphinx build process"):
        # In general we try to avoid passing the "app" object between functions
        # and mutating it at too many places in the stack (although in earlier
        # versions this was the primary idiom). This call is a noted exception,
        # and makes it possible to run portions of this process in separate
        # targets.

        sphinx_publication(conf, app)


# sphinx_publication is its own function because it's called as part of some
# giza.operations.deploy tasks (i.e. ``push``).


def sphinx_publication(conf, app):
    """
    :arg Configuration c: A :class:`giza.config.main.Configuration()` object.

    :arg BuildApp app: A :class:`giza.libgiza.app.BuildApp()` object.

    Adds all required tasks to build a Sphinx site. Specifically:

    1. Iterates through the (language * builder * edition) combination and adds
       tasks to generate the content in the
       <build>/<branch>/source<-edition<-language>> directory. There is one
       version of the <build>/<branch>/source directory for every
       language/edition combination, but multiple builders can use the same
       diretory as needed.

    2. Add a task to run the ``sphinx-build`` task.

    3. Run all tasks in proper order.

    4. Process and print the output of ``sphinx-build``.

    :return: The sum of all return codes from all ``sphinx-build`` tasks. All
             non-zero statuses represent errors.

    :rtype: int
    """
    # call function to prepare the source/content. Seperate so that we can also
    # call this from/as the giza.operations.generate.source() entry point.
    sphinx_content_preperation(app, conf)

    app.randomize = False
    app.run()
    app.reset()

    # task that just runs the sphinx build process and returns the summed error
    # code. in a separate function to make it easier to *just* call these tasks
    # from the CI environment via giza.operations.generate.sphinx target
    return sphinx_builder_tasks(app, conf)


def sphinx_builder_tasks(app, conf):
    for ((edition, language, builder), (build_config, sconf)) in get_builder_jobs(conf):
        sphinx_job = sphinx_tasks(sconf, build_config)
        sphinx_job.finalizers = finalize_sphinx_build(sconf, build_config)

        app.extend_queue(sphinx_job)
        logger.info("adding builder job for {0} ({1}, {2})".format(builder, language, edition))

    logger.debug("sphinx build configured, running the build now.")
    app.run()
    logger.debug("sphinx build complete.")
    logger.info('builds finalized. sphinx output and errors to follow')

    # process the sphinx build. These oeprations allow us to de-duplicate
    # messages between builds.
    results = [o for o in app.results
               if isinstance(o, tuple) and len(o) == 2]

    if len(results) == 0:
        # this happens (rarely) if the deps on the sphinx task do *not* trigger
        # sphinx-build to run.

        output = []
        ret_code = 0
    else:
        # add all builders response codes. If they're all then we can return 0,
        # otherwise, exit.
        ret_code = sum([o[0] for o in results])

        output = [o[1].split('\n') for o in results if o != '']

        sphinx_output = list(reduce(itertools.chain, output))
        try:
            output_sphinx_stream(sphinx_output, conf)
        except:
            logger.error('problem parsing sphinx output, exiting')
            raise SystemExit(1)

        if ret_code != 0:
            raise SystemExit(ret_code)

    return ret_code


def sphinx_content_preperation(app, conf):
    # Download embedded git repositories and then run migrations before doing
    # anything else.
    with app.context() as asset_app:
        asset_app.extend_queue(assets_tasks(conf))

    with app.context() as migration_app:
        migration_app.extend_queue(migration_tasks(conf))

    # Copy all source to the ``build/<branch>/source`` directory.
    with Timer('migrating source to build'):
        with app.context(randomize=True) as source_app:
            for (_, (build_config, sconf)) in get_restricted_builder_jobs(conf):
                source_app.extend_queue(source_tasks(build_config, sconf))

    # load all generated content and create tasks.
    with Timer('loading generated content'):
        for (_, (build_config, sconf)) in get_restricted_builder_jobs(conf):
            for content, func in build_config.system.content.task_generators:
                app.add(Task(job=func,
                             args=[build_config],
                             target=True))

        app.randomize = True
        results = app.run()
        app.reset()

        for task_group in results:
            app.extend_queue(task_group)

    for ((edition, language, builder), (build_config, sconf)) in get_restricted_builder_jobs(conf):
        # these functions all return tasks
        app.extend_queue(image_tasks(build_config, sconf))
        for content_generator in (robots_txt_tasks, intersphinx_tasks,
                                  hash_tasks):
            app.extend_queue(content_generator(build_config))

        dependency_refresh_app = app.add('app')
        dependency_refresh_app.extend_queue(refresh_dependency_tasks(build_config))

        # once the source is prepared, we dump a dict with md5 hashes of all
        # files, so we can do better dependency resolution the next time.
        app.extend_queue(dump_file_hash_tasks(build_config))

        # we transfer images to the latex directory directly because offset
        # images are included using raw latex, and Sphinx doesn't know how
        # to copy images in this case.
        app.extend_queue(latex_image_transfer_tasks(build_config, sconf))

        msg = 'added source tasks for ({0}, {1}, {2}) in {3}'
        logger.debug(msg.format(builder, language, edition, build_config.paths.branch_source))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import shutil

from giza.content.release.inheritance import ReleaseDataCache
from giza.content.release.views import render_releases
from giza.config.content import new_content_type
from giza.libgiza.task import Task

logger = logging.getLogger('giza.content.release.tasks')


def register_releases(conf):
    content_dfn = new_content_type(name='release',
                                   task_generator=release_tasks,
                                   conf=conf)

    conf.system.content.add(name='releases', definition=content_dfn)


def write_release_file(release, fn, conf):
    content = render_releases(release, conf)
    content.write(fn)
    logger.debug('wrote release content: ' + fn)


def release_tasks(conf):
    rel = ReleaseDataCache(conf.system.content.releases.sources, conf)
    rel.create_output_dir()

    tasks = []
    for dep_fn, release in rel.content_iter():
        t = Task(job=write_release_file,
                 args=(release, release.target, conf),
                 description='generating release spec file: ' + release.target,
                 target=release.target,
                 dependency=dep_fn)
        tasks.append(t)

    logger.debug('added tasks for {0} release generation tasks'.format(len(tasks)))
    return tasks


def release_clean(conf):
    return [Task(job=shutil.rmtree,
                 args=[conf.system.content.releases.output_dir],
                 target=True,
                 dependency=[conf.system.content.releases.output_dir],
                 description='removing {0}'.format(conf.system.content.releases.output_dir))]



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import itertools
import operator
import subprocess
import shlex

import yaml

from giza.tools.files import expand_tree


def include_files(conf, files=None):
    if files is not None:
        return files
    else:
        source_dir = os.path.join(conf.paths.projectroot, conf.paths.source)
        cmd = shlex.split('grep -R ".. include:: /" {0}'.format(source_dir))

        with open(os.devnull, 'w') as null:
            try:
                grep = subprocess.check_output(args=cmd, stderr=null)
            except subprocess.CalledProcessError as e:
                grep = e.output

        rx = re.compile(source_dir + r'(.*):.*\.\. include:: (.*)')

        s = [m.groups()
             for m in [rx.match(d)
                       for d in grep.split('\n')]
             if m is not None
             ]

        def tuple_sort(k):
            return k[1]
        s.sort(key=tuple_sort)

        files = dict()

        for i in itertools.groupby(s, operator.itemgetter(1)):
            files[i[0]] = set()
            for src in i[1]:
                if not src[0].endswith('~') and not src[0].endswith('overview.rst'):
                    files[i[0]].add(src[0])
            files[i[0]] = list(files[i[0]])
            files[i[0]].sort()

        for k, v in generated_includes(conf).items():
            if k in files:
                files[k].extend(v)
            else:
                files[k] = v

        return files


def included_once(conf, inc_files=None):
    results = []
    for file, includes in include_files(conf=conf, files=inc_files).items():
        if len(includes) == 1:
            results.append(file)
    return results


def included_recusively(conf, inc_files=None):
    files = include_files(conf=conf, files=inc_files)
    included_files = set(files.keys())

    results = {}
    for inc, srcs in files.items():
        for src in srcs:
            if src in included_files:
                results[inc] = srcs
                break

    return results


def includes_masked(mask, conf, inc_files=None):
    files = include_files(conf=conf, files=inc_files)

    results = {}
    try:
        m = mask + '.rst'
        results[m] = files[m]
    except (ValueError, KeyError):
        for pair in files.items():
            if pair[0].startswith(mask):
                results[pair[0]] = pair[1]

    return results


def generated_includes(conf):
    step_files = []
    mapping = {}

    content_prefixes = []
    for _, prefixes in conf.system.content.content_prefixes:
        content_prefixes.extend(prefixes)

    for fn in expand_tree(os.path.join(conf.paths.includes), input_extension='yaml'):
        base = os.path.basename(fn)

        # example/toc-specs files, for the purpose of this have the same
        # structure as steps, so we can just use that
        for prefix in content_prefixes:
            if base.startswith(prefix):
                step_files.append(fn)
                break

    maskl = len(conf.paths.source)
    path_prefix = conf.paths.includes[len(conf.paths.source):]

    for step_def in step_files:
        deps = []

        with open(step_def, 'r') as f:
            data = yaml.safe_load_all(f)

            for step in data:
                if 'source' in step:
                    deps.append(step['source']['file'])

        if len(deps) != 0:
            deps = [os.path.join(path_prefix, i) for i in deps]

            mapping[step_def[maskl:]] = deps

    return mapping


def include_files_unused(conf, inc_files=None):
    inc_files = [fn[6:] for fn in expand_tree(os.path.join(conf.paths.includes), None)]
    keys = set(include_files(conf=conf).keys())

    results = []
    for fn in inc_files:
        if fn.endswith('yaml') or fn.endswith('~'):
            continue
        if fn not in keys:
            results.append(fn)

    return results


def changed_includes(conf):
    from pygit2 import Repository, GIT_STATUS_CURRENT, GIT_STATUS_IGNORED

    repo_path = conf.paths.projectroot

    r = Repository(repo_path)

    changed = []
    for path, flag in r.status().items():
        if flag not in [GIT_STATUS_CURRENT, GIT_STATUS_IGNORED]:
            if path.startswith('source/'):
                if path.endswith('.txt') or path.endswith('.rst'):
                    changed.append(path[6:])

    changed_report = []

    for fn in include_files(conf):
        if fn in changed:
            changed_report.append(fn)

    return changed_report



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`~giza.pool` provides methods for executing the tasks in a
:class:`~giza.app`. The base class :class:`~giza.pool.WorkerPool` provides core
functionality, while additional sub-classes use different parallelism
mechanisms.
"""

import logging
import multiprocessing
import multiprocessing.dummy
import numbers
import sys

from giza.libgiza.task import MapTask, Task

logger = logging.getLogger('giza.pool')


class PoolConfigurationError(Exception):
    pass


class PoolResultsError(Exception):
    pass


def run_task(task):
    "helper to call run method on task so entire operation can be pickled for process pool support"

    try:
        result = task.run()
    except KeyboardInterrupt:
        logger.error('task received interrupt.')

    return result


class WorkerPool(object):
    @property
    def pool_size(self):
        try:
            return self._pool_size
        except:
            return 2

    @pool_size.setter
    def pool_size(self, value):
        if isinstance(value, numbers.Number):
            self._pool_size = value
        else:
            self._pool_size = multiprocessing.cpu_count()
            logger.error('{0} is not a valid pool size, using number of cores'.format(str(value)))

    def __enter__(self):
        return self.p

    def __exit__(self, *args):
        self.close()

    def close(self):
        self.p.close()
        self.p.join()

    def runner(self, jobs):
        return self.get_results(self.async_runner(jobs))

    def async_runner(self, jobs):
        results = []

        for job in jobs:
            if not hasattr(job, 'run'):
                raise TypeError('task "{0}" is not a valid Task'.format(job))

            if job.needs_rebuild is True:
                self.add_task(job, results)
            else:
                logger.debug("{0} does not need a rebuild".format(job.target))

        return results

    def do_finalizers(self, job, results):
        final = None

        if len(job.finalizers) == 0:
            pass
        else:
            for task in job.finalizers:
                if isinstance(task, tuple) and task[0] in ('final', 'last'):
                    if final is not None:
                        logger.error('can only define one final finalizer task')
                    else:
                        final = task[1]
                else:
                    if task.needs_rebuild is True:
                        self.add_task(task, results)

        self.add_task(final, results)

    def add_task(self, job, results):
        idx = len(results) + 1

        if job is None:
            return
        elif hasattr(job, 'queue'):
            m = 'cannot use finalizers that have queues. skipping tasks ({0})'
            logger.warning(m.format(len(job.queue)))

        if isinstance(job, MapTask):
            results.append((job, idx, self.p.map_async(job.job, job.iter)))
        else:
            results.append((job, idx, self.p.apply_async(run_task, args=[job])))

    def get_results(self, results):
        has_errors = False

        retval = []
        errors = []

        has_finalizers = False
        for job, _, __ in results:
            if len(job.finalizers) >= 1:
                has_finalizers = True
                break

        # the while loop is required so that finalizer tasks begin running as
        # soon as their parent completes, but is otherwise functionally
        # equivalent . However, the while loop spins the main thread, which is
        # *much* slower with thread/gevent pools
        if has_finalizers is True:
            while True:
                for job, idx, ret in results:
                    if ret.ready():
                        try:
                            task_result = ret.get()
                        except Exception as e:
                            if job.ignore_errors is True:
                                m = 'caught error "{0}", waiting for other tasks to finish'
                                logger.error(m.format(e))
                                has_errors = True
                                errors.append((job, e))
                            else:
                                m = "caught error {0} with task {1}. exiting now."
                                logger.error(m.format(e, job.description))
                                raise SystemExit(1)

                        retval.append((idx, task_result))
                        self.do_finalizers(job, results)
                        results.remove((job, idx, ret))

                if len(results) == 0:
                    break

            retval.sort(key=lambda x: x[0])
        else:
            for job, idx, ret in results:
                try:
                    retval.append((idx, ret.get()))
                except Exception as e:
                    if job.ignore_errors is True:
                        m = 'caught error "{0}" in {1}, waiting for other tasks to finish'
                        logger.error(m.format(e, job.description))
                        has_errors = True
                        errors.append((job, e))
                    else:
                        m = "caught error {0} with task {1}. exiting now."
                        logger.error(m.format(e, job.description))
                        raise SystemExit(1)

        if has_errors is True:
            error_list = []
            for job, err in errors:
                error_list.append(err)

            logger.error(PoolResultsError(error_list))
            raise SystemExit(1)
        else:
            return [r[1] for r in retval]


class SerialPool(object):
    def __init__(self, pool_size=0):
        self.p = None
        self.pool_size = pool_size
        logger.debug('new phony "serial" pool object')

    def close(self):
        pass

    def get_results(self, results):
        return results

    def runner(self, jobs):
        results = []
        for job in jobs:
            if job.needs_rebuild is False:
                continue

            if job.description is not None:
                msg = job.description
            else:
                msg = str(job.job)

            logger.debug('running: ' + msg)
            results.append(job.run())

            if isinstance(job, Task) and len(job.finalizers) >= 1:
                logger.debug('finalizing: ' + msg)
                results.extend(job.finalize())

        return results

    async_runner = runner


class ThreadPool(WorkerPool):
    def __init__(self, pool_size=None):
        self.pool_size = pool_size
        self.p = multiprocessing.dummy.Pool(self.pool_size)
        logger.debug('new thread pool object')


class ProcessPool(WorkerPool):
    def __init__(self, pool_size=None):
        self.pool_size = pool_size
        self.p = multiprocessing.Pool(self.pool_size)
        logger.debug('new process pool object')


class EventPool(WorkerPool):
    def __init__(self, pool_size=None):
        self.pool_size = pool_size

        if sys.version_info >= (3, 0):
            logger.error('gevent is not supported on this platform, using threads')
            self.p = multiprocessing.dummy.Pool(self.pool_size)
        else:
            try:
                import gevent.pool
                self.p = gevent.pool.Pool(self.pool_size)
                logger.debug('new event pool object')
            except ImportError:
                logger.error('gevent is not supported on this system, using threads')
                self.p = multiprocessing.dummy.Pool(self.pool_size)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from rstcloth.rstcloth import RstCloth
from giza.content.steps.views import render_action

logger = logging.getLogger('giza.content.extracts.views')


def render_extracts(extract):
    r = RstCloth()
    extract.render()

    indent = 0
    if 'only' in extract:
        r.directive('only', extract.only, indent=indent)
        r.newline()
        indent += 3

    if 'style' in extract:
        r.directive('rst-class', extract.style, indent=indent)
        r.newline()

    render_action(extract, indent=indent, level=extract.level, r=r)

    return r


def get_include_statement(include_file):
    r = RstCloth()
    r.newline()

    r.directive('include', include_file)
    r.newline()

    return '\n'.join(r.data)
